from __future__ import annotations

import typing

import absorb
from .. import cli_outputs
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from argparse import Namespace
    from typing import Any


def sql_command(args: Namespace) -> dict[str, Any]:
    lf = absorb.ops.sql_query(args.sql, backend=args.backend)
    df = lf.collect()
    print(df)

    if args.output_file is not None:
        absorb.ops.write_file(df=df, path=args.output_file)
        print('wrote output to', args.output_file)

    return {}



from __future__ import annotations

import typing

import absorb

if typing.TYPE_CHECKING:
    import polars as pl


class Commits(absorb.Table):
    source = 'git'
    description = 'Commit history of git repository'
    url = 'https://git-scm.com/'
    write_range = 'overwrite_all'
    parameter_types = {'paths': list[str]}
    index_type = 'id'
    require_name = True
    required_packages = ['nitwit >= 1.1']

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'hash': pl.String,
            'author': pl.String,
            'email': pl.String,
            'timestamp': pl.Datetime(time_unit='ms', time_zone='UTC'),
            'message': pl.String,
            'parents': pl.String,
            'committer': pl.String,
            'committer_email': pl.String,
            'commit_timestamp': pl.Datetime(time_unit='ms', time_zone='UTC'),
            'tree_hash': pl.String,
            'is_merge': pl.Boolean,
            'repo_author': pl.String,
            'repo_name': pl.String,
            'repo_source': pl.String,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import nitwit
        import polars as pl

        dfs = [
            nitwit.collect_commits(path) for path in self.parameters['paths']
        ]
        return pl.concat(dfs)


class Authors(absorb.Table):
    source = 'git'
    description = 'Author stats of git repository'
    url = 'https://git-scm.com/'
    write_range = 'overwrite_all'
    parameter_types = {'path': str}
    require_name = True
    index_type = 'id'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'author': pl.String,
            'email': pl.String,
            'first_commit_timestamp': pl.Datetime(
                time_unit='ms', time_zone='UTC'
            ),
            'last_commit_timestamp': pl.Datetime(
                time_unit='ms', time_zone='UTC'
            ),
            'n_commits': pl.Int64,
            'n_repos': pl.UInt32,
            'repo_source': pl.String,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import nitwit

        dfs = []
        for path in self.parameters['paths']:
            commits = nitwit.collect_commits(self.parameters['path'])
            df = nitwit.collect_authors(commits)
            df = df.with_columns(repo_source=pl.lit(path))
            dfs.append(df)
        return pl.concat(dfs)


class FileDiffs(absorb.Table):
    source = 'git'
    description = 'File diffs of git repository'
    url = 'https://git-scm.com/'
    write_range = 'overwrite_all'
    parameter_types = {'path': str}
    require_name = True
    index_type = 'id'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'hash': pl.String,
            'insertions': pl.Int64,
            'deletions': pl.Int64,
            'path': pl.String,
            'repo_author': pl.String,
            'repo_name': pl.String,
            'repo_source': pl.String,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import nitwit

        dfs = [
            nitwit.collect_file_diffs(path) for path in self.parameters['paths']
        ]
        return pl.concat(dfs)


class FileDiffStats(absorb.Table):
    source = 'git'
    description = 'File diff statistics of git repository'
    url = 'https://git-scm.com/'
    write_range = 'overwrite_all'
    parameter_types = {'path': str}
    require_name = True
    index_type = 'id'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'hash': pl.String,
            'n_changed_files': pl.UInt32,
            'insertions': pl.Int64,
            'deletions': pl.Int64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import nitwit

        dfs = [
            nitwit.collect_commit_file_diffs(path).with_columns(
                repo_source=pl.lit(path)
            )
            for path in self.parameters['paths']
        ]
        return pl.concat(dfs)



from __future__ import annotations

import typing

import absorb
from . import cli_helpers

if typing.TYPE_CHECKING:
    import argparse


def get_subcommands() -> list[
    tuple[str, str, list[tuple[list[str], dict[str, typing.Any]]]]
]:
    return [
        (
            'ls',
            'list tracked datasets',
            [
                (
                    ['source'],
                    {
                        'nargs': '?',
                        'help': 'data source',
                    },
                ),
                (
                    ['--available'],
                    {
                        'action': 'store_true',
                        'help': 'list available datasets',
                    },
                ),
                (
                    ['--tracked'],
                    {
                        'action': 'store_true',
                        'help': 'list tracked datasets',
                    },
                ),
                (
                    ['--untracked-collected'],
                    {
                        'action': 'store_true',
                        'help': 'list untracked collected datasets',
                    },
                ),
                (
                    ['--one-per-line', '-1'],
                    {
                        'action': 'store_true',
                        'help': 'list one dataset per line',
                    },
                ),
                (
                    ['--verbose', '-v'],
                    {
                        'action': 'store_true',
                        'help': 'show verbose details',
                    },
                ),
            ],
        ),
        (
            'info',
            'show info about a specific dataset or source',
            [
                (
                    ['dataset_or_source'],
                    {
                        'help': 'dataset or data source',
                    },
                ),
                (
                    ['--verbose', '-v'],
                    {
                        'action': 'store_true',
                        'help': 'show verbose details',
                    },
                ),
            ],
        ),
        (
            'collect',
            'collect datasets',
            [
                (
                    ['dataset'],
                    {
                        'nargs': '*',
                        'help': 'dataset to track, format as "<source>.<dataset>"',
                    },
                ),
                (
                    ['--parameters'],
                    {'nargs': '*', 'help': 'dataset parameters'},
                ),
                (
                    ['--dry'],
                    {
                        'action': 'store_true',
                        'help': 'perform dry run (avoids collecting data)',
                    },
                ),
                (
                    ['--overwrite'],
                    {
                        'action': 'store_true',
                        'help': 'overwrite existing files',
                    },
                ),
                (
                    ['--range'],
                    {
                        'help': 'range of data to collect',
                    },
                ),
                (
                    ['--setup-only'],
                    {
                        'action': 'store_true',
                        'help': 'only setup the table directory, do not collect data',
                    },
                ),
                (
                    ['-v', '--verbose'],
                    {
                        'help': 'display extra information',
                        'nargs': '?',
                        'const': 1,
                        'default': 1,
                        'type': int,
                    },
                ),
            ],
        ),
        (
            'add',
            'start tracking datasets',
            [
                (
                    ['dataset'],
                    {
                        'nargs': '*',
                        'help': 'dataset to track, format as "<source>.<dataset>"',
                    },
                ),
                (
                    ['--parameters'],
                    {
                        'nargs': '*',
                        'help': 'dataset parameters as `key=value` args',
                        'metavar': 'PARAMS',
                    },
                ),
                (
                    ['--path'],
                    {'help': 'directory location to store the dataset'},
                ),
                (
                    ['--collected'],
                    {
                        'action': 'store_true',
                        'help': 'add all datasets that are already collected',
                    },
                ),
            ],
        ),
        (
            'remove',
            'remove tracking datasets',
            [
                (
                    ['dataset'],
                    {
                        'nargs': '*',
                        'help': 'dataset to track, format as "<source>.<dataset>"',
                    },
                ),
                (
                    ['--parameters'],
                    {
                        'nargs': '*',
                        'help': 'dataset parameters',
                        'metavar': 'PARAM=VALUE',
                    },
                ),
                (
                    ['--all'],
                    {
                        'help': 'add all available datasets',
                        'action': 'store_true',
                    },
                ),
                (
                    ['--delete'],
                    {
                        'action': 'store_true',
                        'help': 'delete the dataset files from disk',
                    },
                ),
                (
                    ['--delete-only'],
                    {
                        'action': 'store_true',
                        'help': 'keep tracking table, but delete the dataset files from disk',
                    },
                ),
                (
                    ['--confirm'],
                    {
                        'action': 'store_true',
                        'help': 'confirm the deletion of dataset files',
                    },
                ),
            ],
        ),
        (
            'path',
            'print absorb root path or dataset path',
            [
                (
                    ['dataset'],
                    {
                        'nargs': '?',
                        'help': 'dataset to track, format as "<source>.<dataset>"',
                    },
                ),
                (
                    ['--parameters'],
                    {
                        'nargs': '*',
                        'help': 'dataset parameters',
                        'metavar': 'PARAMS',
                    },
                ),
                (
                    ['--glob'],
                    {'action': 'store_true'},
                ),
            ],
        ),
        (
            'cd',
            'change directory to an absorb path',
            [
                (
                    ['dataset'],
                    {
                        'nargs': '?',
                        'help': 'dataset to track, format as "<source>.<dataset>"',
                    },
                ),
                (
                    ['--parameters'],
                    {
                        'nargs': '*',
                        'help': 'dataset parameters',
                        'metavar': 'PARAMS',
                    },
                ),
                (
                    ['--glob'],
                    {'action': 'store_true'},
                ),
            ],
        ),
        (
            'new',
            'create new dataset',
            [
                (
                    ['dataset'],
                    {
                        'nargs': '?',
                        'help': 'dataset to track, format as "<source>.<dataset>"',
                    },
                ),
                (
                    ['--path'],
                    {
                        'help': 'path where to store new table definition',
                    },
                ),
                (
                    ['--native'],
                    {
                        'action': 'store_true',
                        'help': 'create definition directly in absorb repo',
                    },
                ),
            ],
        ),
        (
            'preview',
            'preview rows of a dataset',
            [
                (
                    ['dataset'],
                    {
                        'nargs': '+',
                        'help': 'dataset to track, format as "<source>.<dataset>"',
                    },
                ),
                (
                    ['--parameters'],
                    {
                        'nargs': '*',
                        'help': 'dataset parameters',
                        'metavar': 'PARAMS',
                    },
                ),
                (
                    ['--count'],
                    {
                        'type': int,
                        'default': 10,
                        'help': 'number of rows to preview',
                    },
                ),
                (
                    ['--offset'],
                    {
                        'type': int,
                        'default': 0,
                        'help': 'number of rows to preview',
                    },
                ),
            ],
        ),
        (
            'sql',
            'run SQL query',
            [
                (
                    ['sql'],
                    {
                        'help': 'SQL query to run',
                    },
                ),
                (
                    ['--backend'],
                    {
                        'default': 'absorb',
                        'help': 'SQL backend to use {absorb, dune, snowflake}',
                    },
                ),
                (
                    ['--output-file'],
                    {
                        'help': 'parquet file to save query results',
                        'nargs': '?',
                        'metavar': 'FILE',
                    },
                ),
            ],
        ),
        (
            'setup',
            'setup environment',
            [
                (
                    ['dataset'],
                    {
                        'nargs': '*',
                        'help': 'dataset to track, format as "<source>.<dataset>"',
                    },
                ),
                (
                    ['--regenerate-metadata'],
                    {
                        'action': 'store_true',
                        'help': 'regenerate metadata for dataset(s)',
                    },
                ),
                (
                    ['--regenerate-config'],
                    {
                        'action': 'store_true',
                        'help': 'regenerate configuration, preserving as many settings as possible',
                    },
                ),
                (
                    ['--enable-git'],
                    {
                        'action': 'store_true',
                        'help': 'enable git tracking for config and metadata',
                    },
                ),
                (
                    ['--disable-git'],
                    {
                        'action': 'store_true',
                        'help': 'disable git tracking for config and metadata',
                    },
                ),
                (
                    ['--set-default-rclone-remote'],
                    {
                        'help': 'set rclone remote to use for bucket uploads and downloads',
                        'metavar': 'REMOTE',
                    },
                ),
                (
                    ['--set-default-bucket'],
                    {
                        'help': 'set default bucket for upload and downloads',
                        'metavar': 'BUCKET',
                    },
                ),
                (
                    ['--set-default-provider'],
                    {
                        'help': 'set default bucket provider for upload and downloads',
                        'metavar': 'PROVIDER',
                    },
                ),
                (
                    ['--set-default-path-prefix'],
                    {
                        'help': 'set default path prefix for bucket paths',
                        'metavar': 'PREFIX',
                    },
                ),
                (
                    ['--clear-default-rclone-remote'],
                    {
                        'help': 'clear rclone remote to use for bucket uploads and downloads',
                        'action': 'store_true',
                    },
                ),
                (
                    ['--clear-default-bucket'],
                    {
                        'help': 'clear bucket to use for upload and downloads',
                        'action': 'store_true',
                    },
                ),
                (
                    ['--clear-default-provider'],
                    {
                        'help': 'clear provider to use for upload and downloads',
                        'action': 'store_true',
                    },
                ),
                (
                    ['--clear-default-path-prefix'],
                    {
                        'help': 'clear default path prefix for bucket paths',
                        'action': 'store_true',
                    },
                ),
                (
                    ['-v', '--verbose'],
                    {
                        'action': 'store_true',
                        'help': 'display extra information',
                    },
                ),
            ],
        ),
        (
            'validate',
            'validate datasets',
            [
                (
                    ['dataset'],
                    {
                        'nargs': '?',
                        'help': 'dataset to track, format as "<source>.<dataset>"',
                    },
                ),
                (
                    ['--parameters'],
                    {
                        'nargs': '*',
                        'help': 'dataset parameters',
                        'metavar': 'PARAMS',
                    },
                ),
                (
                    ['--verbose', '-v'],
                    {
                        'action': 'store_true',
                        'help': 'display extra information',
                    },
                ),
            ],
        ),
        (
            'upload',
            'upload datasets to a cloud bucket',
            [
                (
                    ['dataset'],
                    {
                        'nargs': '*',
                        'help': 'dataset to track, format as "<source>.<dataset>"',
                    },
                ),
                (
                    ['--parameters'],
                    {
                        'nargs': '*',
                        'help': 'dataset parameters',
                        'metavar': 'PARAMS',
                    },
                ),
                (
                    ['--rclone-remote'],
                    {
                        'help': 'name of rclone remote to use',
                        'metavar': 'REMOTE',
                    },
                ),
                (
                    ['--bucket'],
                    {
                        'help': 'name of bucket to upload to',
                        'metavar': 'BUCKET',
                    },
                ),
                (
                    ['--path-prefix'],
                    {
                        'help': 'path prefix to use for the bucket',
                        'metavar': 'PREFIX',
                    },
                ),
                (
                    ['--dry'],
                    {
                        'action': 'store_true',
                        'help': 'perform dry run (avoids uploading data)',
                    },
                ),
            ],
        ),
        (
            'download',
            'download datasets to a cloud bucket',
            [
                (
                    ['dataset'],
                    {
                        'nargs': '*',
                        'help': 'dataset to track, format as "<source>.<dataset>"',
                    },
                ),
                (
                    ['--parameters'],
                    {
                        'nargs': '*',
                        'help': 'dataset parameters',
                        'metavar': 'PARAMS',
                    },
                ),
                (
                    ['--rclone-remote'],
                    {
                        'help': 'name of rclone remote to use',
                        'metavar': 'REMOTE',
                    },
                ),
                (
                    ['--bucket'],
                    {
                        'help': 'name of bucket to download to',
                        'metavar': 'BUCKET',
                    },
                ),
                (
                    ['--path-prefix'],
                    {
                        'help': 'path prefix to use for the bucket',
                        'metavar': 'PREFIX',
                    },
                ),
                (
                    ['--dry'],
                    {
                        'action': 'store_true',
                        'help': 'perform dry run (avoids downloading data)',
                    },
                ),
            ],
        ),
    ]


def get_common_args() -> list[tuple[list[str], dict[str, typing.Any]]]:
    import argparse

    return [
        (
            [
                '--debug',
                '--pdb',
            ],
            {
                'help': 'enter debugger upon error',
                'action': 'store_true',
            },
        ),
        (
            [
                '-i',
                '--interactive',
            ],
            {
                # 'help': 'open data in interactive python session',
                'help': argparse.SUPPRESS,
                'action': 'store_true',
            },
        ),
        (
            [
                '--absorb-root',
            ],
            {
                'help': 'path to absorb root directory',
                'metavar': 'PATH',
            },
        ),
        (
            [
                '--cd-destination-tempfile',
            ],
            {
                'help': argparse.SUPPRESS,
            },
        ),
    ]


def parse_args() -> argparse.Namespace:
    """parse input arguments into a Namespace object"""
    import argparse
    import importlib
    import sys

    # create top-level parser
    parser = argparse.ArgumentParser(
        formatter_class=cli_helpers.HelpFormatter, allow_abbrev=False
    )
    parser.add_argument('--cd-destination-tempfile', help=argparse.SUPPRESS)

    # create subparsers
    subparsers = parser.add_subparsers(dest='command')
    common_args = get_common_args()
    for name, description, arg_args in get_subcommands():
        module_name = 'absorb.cli.cli_commands.command_' + name
        f_module = importlib.import_module(module_name)
        subparser = subparsers.add_parser(name, help=description)
        subparser.set_defaults(f_command=getattr(f_module, name + '_command'))
        for sub_args, sub_kwargs in arg_args + common_args:
            subparser.add_argument(*sub_args, **sub_kwargs)

    # parse args
    args = parser.parse_args()

    # display help if no command specified
    if args.command is None:
        parser.print_help()
        sys.exit(0)

    return args


def _parse_datasets(args: argparse.Namespace) -> list[absorb.Table]:
    """parse the datasets parameter into a list of instantiated Tables"""
    # parse parameters
    parameters: dict[str, typing.Any] = {}
    if args.parameters is not None:
        for parameter in args.parameters:
            key, value = parameter.split('=')
            parameters[key] = value

    # parse tables
    tables = []
    for table_str in args.dataset:
        table = absorb.Table.instantiate(table_str, raw_parameters=parameters)
        tables.append(table)

    return tables


def _parse_ranges(
    raw_ranges: list[str] | None, index_type: absorb.IndexType
) -> list[typing.Any] | None:
    """
    examples:
    --range 2025-01-01:2025-03-01
    --range 2025-01-01:
    --range :2025-01-01
    """
    if raw_ranges is None:
        return None
    if index_type == 'day' or index_type == 'timestamp_range':
        raise NotImplementedError('manual ranges for ' + str(index_type))
    else:
        raise NotImplementedError('manual ranges for ' + str(index_type))

    # 'date',
    # 'date_range',
    # 'named_range',
    # 'block_range',
    # 'id_range',
    # 'count',
    # None,


def _parse_bucket(args: argparse.Namespace) -> absorb.Bucket:
    default_bucket = absorb.ops.get_config()['default_bucket']
    if args.rclone_remote is not None:
        rclone_remote = args.rclone_remote
    else:
        rclone_remote = default_bucket['rclone_remote']
    if args.path_prefix is not None:
        path_prefix = args.path_prefix
    else:
        path_prefix = default_bucket['path_prefix']
    if args.bucket is not None:
        bucket_name = args.bucket
    else:
        bucket_name = default_bucket['bucket_name']
    if args.provider is not None:
        provider = args.provider
    else:
        provider = default_bucket['provider']
    return {
        'rclone_remote': rclone_remote,
        'bucket_name': bucket_name,
        'path_prefix': path_prefix,
        'provider': provider,
    }



from __future__ import annotations

from typing_extensions import NotRequired
import typing
import datetime
import types

import polars as pl

from . import table


# chunk formats
PrimitiveIndexType = typing.Literal[
    # temporal
    'hour',
    'day',
    'week',
    'month',
    'quarter',
    'year',
    'timestamp',
    'timestamp_range',
    # numerical
    'number',
    'number_range',
    'number_list',
    # names
    'id',
    'id_list',
]


class CustomIndexType:
    def partition_into_chunks(self, coverage: Coverage) -> list[Chunk]:
        raise NotImplementedError()

    def format_value(self, chunk: CustomChunk) -> str:
        raise NotImplementedError()


class MultiIndexType(typing.TypedDict):
    type: typing.Literal['multi']
    dims: dict[str, ScalarIndexType]


class ExplicitIndexType(typing.TypedDict):
    type: PrimitiveIndexType
    number_interval: NotRequired[int | None]
    timestamp_interval: NotRequired[datetime.timedelta | None]


ScalarIndexType = typing.Union[
    PrimitiveIndexType,
    ExplicitIndexType,
    CustomIndexType,
]

IndexType = typing.Union[ScalarIndexType, MultiIndexType]

# chunks
PrimitiveChunk = typing.Union[
    datetime.datetime,
    tuple[datetime.datetime, datetime.datetime],
    int,
    list[int],
    tuple[int, int],
    str,
    list[str],
    tuple[str, str],
]
CustomChunk = typing.Any
ScalarChunk = typing.Union[PrimitiveChunk, CustomChunk]
MultiChunk = dict[str, ScalarChunk]
Chunk = typing.Union[ScalarChunk, MultiChunk]

# chunk coverage
ChunkList = typing.Sequence[Chunk]
ScalarChunkRange = tuple[ScalarChunk, ScalarChunk]
MultiChunkRange = typing.Mapping[
    str, typing.Union[ScalarChunkRange, list[ScalarChunk]]
]
Coverage = typing.Union[ChunkList, ScalarChunkRange, MultiChunkRange]


class ChunkPaths(typing.TypedDict):
    type: typing.Literal['files']
    paths: list[str]


ChunkData = typing.Union[pl.DataFrame, ChunkPaths]


def get_index_type_type(
    index_type: PrimitiveIndexType,
) -> type | types.GenericAlias | None:
    import datetime

    if isinstance(index_type, str):
        return {
            'hour': datetime.datetime,
            'day': datetime.datetime,
            'week': datetime.datetime,
            'month': datetime.datetime,
            'quarter': datetime.datetime,
            'year': datetime.datetime,
            'timestamp': datetime.datetime,
            'timestamp_range': tuple[datetime.datetime, datetime.datetime],
            'count': int,
            'count_range': tuple[int, int],
            'id': str,
            'id_list': list[str],
            None: None,
        }[index_type]
    elif isinstance(index_type, dict):
        return dict[str, typing.Any]
    else:
        raise Exception()


JSONValue = typing.Union[
    str,
    int,
    float,
    bool,
    None,
    dict[str, 'JSONValue'],
    list['JSONValue'],
]


class TableDict(typing.TypedDict):
    source_name: str
    table_name: str
    table_class: str
    parameters: dict[str, JSONValue]
    table_version: str


TableReference = typing.Union[
    str,
    tuple[str, dict[str, JSONValue]],
    TableDict,
    table.Table,
]


class Bucket(typing.TypedDict):
    rclone_remote: str | None
    bucket_name: str | None
    path_prefix: str | None
    provider: str | None


class Config(typing.TypedDict):
    version: str
    tracked_tables: list[TableDict]
    use_git: bool
    default_bucket: Bucket



from __future__ import annotations

import absorb
from . import cli_parsing
from . import cli_helpers


def run_cli() -> None:
    args = cli_parsing.parse_args()

    if args.absorb_root is not None:
        absorb.ops.set_absorb_root(args.absorb_root)

    try:
        data = args.f_command(args)
        if args.interactive:
            cli_helpers.open_interactive_session(variables=data)
    except BaseException as e:
        if isinstance(e, SystemExit) and len(e.args) > 0 and e.args[0] == 0:
            pass
        elif args.debug:
            cli_helpers._enter_debugger()
        else:
            import toolstr

            toolstr.print('[red]error:[/red] ' + str(e.args[0]))



from __future__ import annotations

import typing

import absorb
from .. import cli_outputs

if typing.TYPE_CHECKING:
    from argparse import Namespace
    from typing import Any


def ls_command(args: Namespace) -> dict[str, Any]:
    import toolstr

    # decide which sections to print
    sections = set()
    if not args.available and not args.tracked and not args.untracked_collected:
        sections.add('available')
        sections.add('tracked')
        sections.add('untracked_collected')
    if args.available:
        sections.add('available')
    if args.tracked:
        sections.add('tracked')
    if args.untracked_collected:
        sections.add('untracked_collected')

    # available datasets
    if 'available' in sections:
        cli_outputs._print_title('Available datasets')
        if args.source is not None:
            sources = [args.source]
        else:
            sources = absorb.ops.get_sources()
        for source in sorted(sources):
            table_classes = absorb.ops.get_source_table_classes(source)
            if args.one_per_line:
                for dataset in table_classes:
                    toolstr.print_bullet(
                        '[white bold]'
                        + dataset.source
                        + '.'
                        + dataset.name_classmethod(allow_generic=True)
                        + '[/white bold]',
                        **absorb.ops.bullet_styles,
                    )
            else:
                if len(table_classes) > 0:
                    names = [
                        cls.name_classmethod(allow_generic=True)
                        for cls in table_classes
                    ]
                    toolstr.print_bullet(
                        key=source,
                        value='[green],[/green] '.join(names),
                        **absorb.ops.bullet_styles,
                    )

    # get tracked datasets
    tracked_datasets = absorb.ops.get_tracked_tables()
    if args.source is not None:
        tracked_datasets = [
            dataset
            for dataset in tracked_datasets
            if dataset['source_name'] == args.source
        ]
    tracked_datasets = sorted(
        tracked_datasets, key=lambda x: (x['source_name'], x['table_name'])
    )

    # print tracked datasets
    if 'tracked' in sections:
        if 'available' in sections:
            print()
        _print_tracked_datasets(
            tracked_datasets,
            verbose=args.verbose,
            one_per_line=args.one_per_line,
        )

    # get untracked collected datasets
    if 'untracked_collected' in sections:
        _print_untracked_datasets(
            tracked_datasets=tracked_datasets,
            source=args.source,
            verbose=args.verbose,
            one_per_line=args.one_per_line,
            skip_line='tracked' in sections or 'available' in sections,
        )

    return {}


def _print_tracked_datasets(
    tracked_datasets: list[absorb.TableDict],
    verbose: bool = False,
    one_per_line: bool = False,
) -> None:
    cli_outputs._print_title(
        'Tracked datasets (n = {})'.format(len(tracked_datasets))
    )
    if len(tracked_datasets) == 0:
        print('[none]')
    else:
        _print_datasets(
            tracked_datasets, verbose=verbose, one_per_line=one_per_line
        )


def _print_untracked_datasets(
    tracked_datasets: list[absorb.TableDict],
    source: str | None = None,
    skip_line: bool = False,
    verbose: bool = False,
    one_per_line: bool = False,
) -> None:
    untracked_collected_datasets = absorb.ops.get_untracked_collected_tables(
        tracked_datasets=tracked_datasets
    )
    if source is not None:
        untracked_collected_datasets = [
            dataset
            for dataset in untracked_collected_datasets
            if dataset['source_name'] == source
        ]
    untracked_collected_datasets = sorted(
        untracked_collected_datasets,
        key=lambda x: (x['source_name'], x['table_name']),
    )

    # print untracked collected datasets
    if len(untracked_collected_datasets) > 0:
        if skip_line:
            print()
        cli_outputs._print_title(
            'Untracked collected datasets (n = {})'.format(
                len(untracked_collected_datasets)
            )
        )
        _print_datasets(
            untracked_collected_datasets,
            verbose=verbose,
            one_per_line=one_per_line,
        )


def _print_datasets(
    datasets: list[absorb.TableDict], verbose: bool, one_per_line: bool
) -> None:
    import toolstr

    if verbose:
        _print_datasets_verbose(datasets)
    elif one_per_line:
        for dataset in datasets:
            cli_outputs._print_dataset_bullet(dataset)
    else:
        tracked_sources = sorted(
            {dataset['source_name'] for dataset in datasets}
        )
        for source in tracked_sources:
            names = []
            for dataset in datasets:
                if dataset['source_name'] == source:
                    try:
                        instance = absorb.Table.instantiate(dataset)
                        name = instance.name()
                    except Exception:
                        name = (
                            dataset['source_name'] + '.' + dataset['table_name']
                        )
                    names.append(name)

            toolstr.print_bullet(
                key=source,
                value='[green],[/green] '.join(names),
                **absorb.ops.bullet_styles,
            )


def _print_datasets_verbose(datasets: list[absorb.TableDict]) -> None:
    import toolstr

    rows = []
    for dataset in datasets:
        instance = absorb.Table.instantiate(dataset)
        available_range = instance.get_available_range()
        available_range_str = absorb.ops.format_coverage(
            available_range, instance.index_type
        )
        collected_range = instance.get_collected_range()
        if collected_range is not None:
            collected_range_str = absorb.ops.format_coverage(
                collected_range, instance.index_type
            )
        else:
            collected_range_str = '-'
        row = [
            dataset['source_name'],
            dataset['table_name'],
            available_range_str + '\n' + collected_range_str,
        ]
        rows.append(row)
    columns = ['source', 'table', 'available range\ncollected range']
    toolstr.print_multiline_table(rows, labels=columns)



from __future__ import annotations

import typing
import absorb

from . import command_ls

if typing.TYPE_CHECKING:
    import argparse


def info_command(args: argparse.Namespace) -> dict[str, typing.Any]:
    if args.dataset_or_source is None:
        import sys

        print('specify dataset to print info')
        sys.exit(0)

    if '.' in args.dataset_or_source:
        return print_dataset_info(
            table_str=args.dataset_or_source, verbose=args.verbose
        )
    else:
        return print_source_info(
            source=args.dataset_or_source, verbose=args.verbose
        )


def print_source_info(source: str, verbose: bool) -> dict[str, typing.Any]:
    import toolstr

    toolstr.print_text_box(
        'Data source = ' + source, style='green', text_style='bold white'
    )
    classes = absorb.ops.get_source_table_classes(source)
    if len(classes) == 1:
        print(str(len(classes)), 'table recipe:')
    else:
        print(str(len(classes)), 'table recipes:')
    for cls in classes:
        toolstr.print_bullet(
            key='[white bold]'
            + cls.source
            + '.'
            + cls.name_classmethod(allow_generic=True)
            + '[/white bold]',
            value=str(cls.description),
            **absorb.ops.bullet_styles,
        )

    tracked_datasets = absorb.ops.get_tracked_tables()
    if source is not None:
        tracked_datasets = [
            dataset
            for dataset in tracked_datasets
            if dataset['source_name'] == source
        ]

    print()
    command_ls._print_tracked_datasets(
        tracked_datasets, verbose=verbose, one_per_line=True
    )
    # print()
    command_ls._print_untracked_datasets(
        tracked_datasets,
        verbose=verbose,
        one_per_line=True,
        source=source,
        skip_line=False,
    )

    return {}


def print_dataset_info(table_str: str, verbose: bool) -> dict[str, typing.Any]:
    import toolstr

    try:
        table = absorb.Table.instantiate(table_str)
        print_table_info(table, verbose=verbose)
    except Exception:
        source, name = table_str.split('.')
        for cls in absorb.ops.get_source_table_classes(source):
            if cls.name_classmethod(
                allow_generic=True
            ) == name or name == absorb.ops.names._camel_to_snake(
                cls.__qualname__
            ):
                return print_recipe_info(cls, verbose=verbose)
        else:
            import sys

            print('could not find match')
            sys.exit(1)

    return {}


def print_recipe_info(
    cls: type[absorb.Table], verbose: bool
) -> dict[str, typing.Any]:
    import toolstr

    toolstr.print_text_box(
        'Table recipe = ' + cls.name_classmethod(allow_generic=True),
        style='green',
        text_style='bold white',
    )
    for attr in [
        'description',
        'url',
        'source',
        'write_range',
        'index_type',
    ]:
        if hasattr(cls, attr):
            value = getattr(cls, attr)
        else:
            value = None
        absorb.ops.print_bullet(key=attr, value=value)

    # parameters
    print()
    toolstr.print('[green bold]parameters[/green bold]')
    if len(cls.parameter_types) == 0:
        print('- [none]')
    else:
        for key, value in cls.parameter_types.items():
            if key in cls.default_parameters:
                default = (
                    ' \\[default = ' + str(cls.default_parameters[key]) + ']'
                )
            else:
                default = ''
            absorb.ops.print_bullet(key=key, value=str(value) + default)

    return {}


def print_table_info(
    table: absorb.Table, verbose: bool
) -> dict[str, typing.Any]:
    import toolstr

    schema = table.get_schema()

    toolstr.print_text_box(
        'dataset = ' + table.name(),
        style='green',
        text_style='bold white',
    )

    for attr in [
        'description',
        'url',
        'source',
        'write_range',
        'index_type',
    ]:
        if hasattr(table, attr):
            value = getattr(table, attr)
        else:
            value = None
        absorb.ops.print_bullet(key=attr, value=value)

    # parameters
    print()
    toolstr.print('[green bold]parameters[/green bold]')
    if table.parameters is None or len(table.parameter_types) == 0:
        print('- [none]')
    else:
        for key, value in table.parameter_types.items():
            if key in table.default_parameters:
                default = (
                    ' \\[default = ' + str(table.default_parameters[key]) + ']'
                )
            else:
                default = ''
            absorb.ops.print_bullet(key=key, value=str(value) + default)

    # schema
    print()
    toolstr.print('[green bold]schema[/green bold]')
    for key, value in schema.items():
        absorb.ops.print_bullet(key=key, value=str(value))

    # collection status
    print()
    toolstr.print('[green bold]status[/green bold]')
    # absorb.ops.print_bullet(key='tracked', value=table.is_tracked())

    if verbose:
        available_range = table.get_available_range()
        if available_range is not None:
            formatted_available_range = absorb.ops.format_coverage(
                available_range, table.index_type
            )
        else:
            formatted_available_range = 'not available'
        absorb.ops.print_bullet(
            key='available range',
            value=formatted_available_range,
        )

    collected_range = table.get_collected_range()
    absorb.ops.print_bullet(
        key='collected range',
        value=absorb.ops.format_coverage(collected_range, table.index_type),
    )

    import os

    path = table.get_table_dir()
    if os.path.isdir(path):
        bytes_str = absorb.ops.format_bytes(absorb.ops.get_dir_size(path))
    else:
        path = '[not collected]'
        bytes_str = '[not collected]'
    absorb.ops.print_bullet(key='path', value=path)
    absorb.ops.print_bullet(key='size', value=bytes_str)

    return {}



from __future__ import annotations

import typing

import absorb
from . import table_coverage


class TableCreate(table_coverage.TableCoverage):
    def __init__(self, parameters: dict[str, absorb.JSONValue] | None = None):
        # set parameters
        if hasattr(type(self), 'parameters'):
            raise Exception(
                'parameters should not be set at the class level, use cls.default_parameters'
            )
        if parameters is None:
            parameters = {}
        else:
            parameters = parameters.copy()

        # set default parameters
        for key, value in self.default_parameters.items():
            parameters.setdefault(key, value)

        # make sure that parameters match the parameter types
        if set(parameters.keys()) != set(self.parameter_types.keys()):
            raise Exception(
                self.full_name()
                + ': parameters must match parameter_types spec'
            )
        self.parameters = parameters

        # make sure that all required parameters are set
        required_parameters: list[str] = []
        for parameter in required_parameters:
            if not hasattr(self, parameter) or getattr(self, parameter) is None:
                raise Exception('missing table parameter: ' + str(parameter))

        # make sure that append only tables have an index type
        if self.write_range == 'append_only' and (
            not hasattr(self, 'index_type') or self.index_type is None
        ):
            raise Exception('index_type is required for append only tables')

        # make sure that table name is valid
        if not absorb.ops.is_valid_name(self.name()):
            raise Exception(f'table name "{self.name()}" is not valid')

        # make sure that source name is valid
        if not absorb.ops.is_valid_name(self.source):
            raise Exception(f'source name "{self.source}" is not valid')

    @staticmethod
    def instantiate(
        ref: absorb.TableReference,
        *,
        raw_parameters: dict[str, str] | None = None,
        use_all_parameters: bool = True,
    ) -> absorb.Table:
        if isinstance(ref, absorb.Table):
            # reference already instantiated
            if raw_parameters is not None:
                raise Exception('Cannot pass parameters with table instance')
            return ref
        elif isinstance(ref, dict):
            # reference is a table dict
            cls = absorb.ops.get_table_class(class_path=ref['table_class'])
            if raw_parameters is not None:
                raise Exception('Cannot pass parameters with table dict')
            parameters = ref['parameters']
        elif isinstance(ref, str):
            # reference is a str
            cls, parameters = absorb.ops.parse_table_str(
                ref,
                raw_parameters=raw_parameters,
                use_all_parameters=use_all_parameters,
            )
        elif isinstance(ref, tuple):
            tuple_name, tuple_parameters = ref
            cls, parameters = absorb.ops.parse_table_str(
                tuple_name,
                parameters=tuple_parameters,
                raw_parameters=raw_parameters,
                use_all_parameters=use_all_parameters,
            )
        else:
            raise Exception()

        return cls(parameters=parameters)



from __future__ import annotations

import typing

import absorb
from .. import cli_outputs
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from argparse import Namespace
    from typing import Any


def remove_command(args: Namespace) -> dict[str, Any]:
    if args.all:
        tracked_datasets = [
            absorb.Table.instantiate(table)
            for table in absorb.ops.get_tracked_tables()
        ]
    else:
        tracked_datasets = cli_parsing._parse_datasets(args)

    if not args.delete_only:
        absorb.ops.stop_tracking_tables(tracked_datasets)
        cli_outputs._print_title('Stopped tracking')
        for dataset in tracked_datasets:
            cli_outputs._print_dataset_bullet(dataset)

    if args.delete:
        print()
        for dataset in tracked_datasets:
            print('deleting files of ' + dataset.full_name())
            try:
                absorb.ops.delete_table_dir(dataset, confirm=args.confirm)
            except absorb.ConfirmError:
                import sys
                import toolstr

                toolstr.print('[red]use --confirm to delete files[/red]')
                sys.exit(0)
        print('...done')
    else:
        print()
        print('to delete table data files, use the --delete flag')

    return {}



from __future__ import annotations

import typing

import absorb

if typing.TYPE_CHECKING:
    import polars as pl


class FourbyteDatatype(absorb.Table):
    source = 'fourbyte'
    url = 'https://www.4byte.directory/'
    write_range = 'append_only'
    index_type = 'number_range'

    # custom
    endpoint: str

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'id': pl.Int64,
            'created_at': pl.Datetime,
            'text_signature': pl.String,
            'hex_signature': pl.String,
            'bytes_signature': pl.Binary,
        }

    def get_available_range(self) -> absorb.Coverage:
        import requests

        data = requests.get(self.endpoint).json()
        max_id = max(result['id'] for result in data['results'])
        return (0, max_id)

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        return scrape_4byte(url=self.endpoint, chunk=chunk)  # type: ignore


class Functions(FourbyteDatatype):
    description = 'Functions listed on 4byte.directory'
    endpoint = 'https://www.4byte.directory/api/v1/signatures/'


class Events(FourbyteDatatype):
    description = 'Event types listed on 4byte.directory'
    endpoint = 'https://www.4byte.directory/api/v1/event-signatures/'


def get_tables() -> list[type[absorb.Table]]:
    return [Functions, Events]


def scrape_4byte(
    url: str,
    chunk: tuple[int, int],
    wait_time: float = 0.1,
    min_id: int | None = None,
) -> pl.DataFrame:
    import requests
    import polars as pl
    import time

    results = []
    while True:
        # get page
        response = requests.get(url)
        result: dict[str, typing.Any] = response.json()
        results.extend(result['results'])

        # scrape only until min_id is reached
        if min_id is not None:
            min_result_id = min(result['id'] for result in result['results'])
            if min_result_id < min_id:
                break

        # get next url
        url = result['next']
        if url is None:
            break

        # wait between responses
        if wait_time is not None:
            time.sleep(wait_time)

    return pl.DataFrame(results, orient='row')



from __future__ import annotations

import typing

import absorb

if typing.TYPE_CHECKING:
    from argparse import Namespace
    from typing import Any


def new_command(args: Namespace) -> dict[str, Any]:
    import os
    import subprocess

    # parse names
    source_name, table_name = args.dataset.split('.')
    snake_table_name = absorb.ops.names._camel_to_snake(table_name)
    camel_table_name = absorb.ops.names._snake_to_camel(table_name)

    # create template content
    table_template = get_table_class_template()
    table_content = table_template.replace(
        '{source_name}', source_name
    ).replace('{class_name}', camel_table_name)

    # determine path
    if args.path is not None:
        path = args.path
    elif args.native:
        source_path = os.path.join(absorb.__path__[0], 'datasets', source_name)
        if os.path.isdir(source_path):
            path = os.path.join(source_path, snake_table_name + '.py')
        else:
            path = os.path.join(source_path + '.py')
    else:
        raise NotImplementedError(
            'creating table definitions in ABSORB_ROOT, use --native or --path'
        )
    print('adding table to file: ' + path)

    # edit file
    os.makedirs(os.path.dirname(path), exist_ok=True)
    if os.path.isfile(path):
        content = '\n\n' + table_content
        with open(path, 'a') as f:
            f.write(content)
    else:
        content = get_data_source_template() + '\n\n\n' + table_content
        with open(path, 'w') as f:
            f.write(content)

    # edit file if EDITOR set
    editor = get_editor()
    if editor is not None:
        subprocess.call([editor, path])
    else:
        print(
            '(not opening editor because EDITOR environment variable not set)'
        )

    return {}


def get_data_source_template() -> str:
    return """from __future__ import annotations

import typing

import absorb

if typing.TYPE_CHECKING:
    import polars as pl"""


def get_table_class_template() -> str:
    template = """class {class_name}(absorb.Table):
    source = '{source_name}'
    write_range = 'overwrite_all'
    parameter_types = {}

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {}

    def collect_chunk(self, chunk: absorb.Chunk) -> pl.DataFrame | None:
        raise NotImplementedError()

    def get_available_range(self) -> absorb.Coverage:
        raise NotImplementedError()
"""
    return template


def get_editor() -> str | None:
    import os

    return os.environ.get('EDITOR')



from __future__ import annotations

import typing

import absorb
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from argparse import Namespace
    from typing import Any


def preview_command(args: Namespace) -> dict[str, Any]:
    import polars as pl
    import toolstr

    preview_length = args.count
    offset = args.offset

    pl.Config.set_tbl_hide_dataframe_shape(True)
    pl.Config.set_tbl_rows(preview_length)

    datasets = cli_parsing._parse_datasets(args)
    n_rows = {}
    for d, dataset in enumerate(datasets):
        # load dataset preview
        df = (
            absorb.query(dataset)
            .slice(offset)
            .head(preview_length + 1)
            .collect()
        )

        # print number of rows in preview
        if d > 0:
            print()
        if len(datasets) > 1:
            toolstr.print_text_box(dataset.full_name(), style='bold')

        if len(df) > preview_length:
            if offset > 0:
                print(preview_length, 'rows starting from offset', offset)
            else:
                print('first', preview_length, 'rows:')

        # print dataset preview
        print(df.head(preview_length))

        # print total number of rows
        dataset_n_rows = absorb.scan(dataset).select(pl.len()).collect().item()
        n_rows[dataset.full_name()] = dataset_n_rows
        print(dataset_n_rows, 'rows,', len(df.columns), 'columns')

    # load interactive previews
    if args.interactive:
        if len(datasets) == 1:
            dataset = datasets[0]
            if n_rows[dataset.full_name()] <= 1_000_000:
                return {'df': absorb.load(dataset)}
            else:
                return {'lf': absorb.scan(dataset)}
        else:
            dfs = {}
            lfs = {}
            for dataset in datasets:
                table_name = dataset.full_name()
                if n_rows[table_name] <= 1_000_000:
                    dfs[table_name] = absorb.load(dataset)
                else:
                    lfs[table_name] = absorb.scan(dataset)
            outputs: dict[str, typing.Any] = {}
            if len(dfs) > 0:
                outputs['dfs'] = dfs
            if len(lfs) > 0:
                outputs['lfs'] = lfs
            return outputs
    else:
        return {}



from __future__ import annotations

import absorb


def _print_title(title: str) -> None:
    import rich

    rich.print('[bold green]' + title + '[/bold green]')


def _print_dataset_bullet(dataset: absorb.TableReference) -> None:
    import toolstr

    table_dict = absorb.Table.instantiate(dataset).create_table_dict()
    full_name = table_dict['source_name'] + '.' + table_dict['table_name']
    toolstr.print_bullet(
        '[white bold]' + full_name + '[/white bold]',
        **absorb.ops.bullet_styles,
    )



from __future__ import annotations

import typing

import absorb


_cache = {'root_dir_warning_shown': False}


def get_absorb_root(*, warn: bool = False) -> str:
    import os

    path = os.environ.get('ABSORB_ROOT')
    if path is None or path == '':
        if warn and not _cache['root_dir_warning_shown']:
            import rich

            rich.print(
                '[#777777]using default value for ABSORB_ROOT: ~/absorb\n(set a value for the ABSORB_ROOT env var to remove this message)[/#777777]'
            )
            _cache['root_dir_warning_shown'] = True
        path = '~/absorb'
    path = os.path.expanduser(path)
    return path


def set_absorb_root(path: str) -> None:
    import os

    os.environ['ABSORB_ROOT'] = path
    _cache['root_dir_warning_shown'] = False


def get_config_path(*, warn: bool = False) -> str:
    import os

    return os.path.join(
        absorb.ops.get_absorb_root(warn=warn), 'absorb_config.json'
    )


def get_datasets_dir(*, warn: bool = False) -> str:
    import os

    return os.path.join(get_absorb_root(warn=warn), 'datasets')


def get_source_tables_dir(source: str, *, warn: bool = False) -> str:
    import os

    return os.path.join(get_datasets_dir(warn=warn), source, 'tables')


def get_source_dir(source: str, *, warn: bool = False) -> str:
    import os

    return os.path.join(get_datasets_dir(warn=warn), source)


def get_table_dir(
    table: str | absorb.TableDict | absorb.Table,
    *,
    source: str | None = None,
    warn: bool = False,
) -> str:
    import os

    if isinstance(table, str):
        if '.' in table:
            source, table = table.split('.')
        else:
            if source is None:
                raise Exception('source must be provided if table is a string')
    elif isinstance(table, dict):
        source = table['source_name']
        table = table['table_name']
    elif isinstance(table, absorb.Table):
        source = table.source
        table = table.name()
    else:
        raise Exception('invalid format')

    source_dir = get_source_dir(source, warn=warn)
    return os.path.join(source_dir, 'tables', table)


def get_table_metadata_path(
    table: str | absorb.TableDict | absorb.Table,
    *,
    source: str | None = None,
    warn: bool = False,
) -> str:
    import os

    table_dir = absorb.ops.get_table_dir(table, source=source, warn=warn)
    return os.path.join(table_dir, 'table_metadata.json')


def get_table_filepath(
    chunk: absorb.Chunk,
    index_type: absorb.IndexType | None,
    filename_template: str,
    table: str,
    *,
    source: str | None,
    parameters: dict[str, typing.Any],
    glob: bool = False,
    warn: bool = True,
) -> str:
    import os

    dir_path = get_table_dir(source=source, table=table, warn=warn)
    filename = get_table_filename(
        chunk=chunk,
        index_type=index_type,
        filename_template=filename_template,
        table=table,
        source=source,
        parameters=parameters,
        glob=glob,
    )
    return os.path.join(dir_path, filename)


def get_table_filename(
    chunk: absorb.Chunk,
    index_type: absorb.IndexType | None,
    filename_template: str,
    table: str,
    *,
    source: str | None,
    parameters: dict[str, typing.Any],
    glob: bool = False,
) -> str:
    format_params = parameters.copy()
    if source is not None:
        format_params['source'] = source
    format_params['table'] = table
    if '{chunk}' in filename_template:
        if glob:
            format_params['chunk'] = '*'
        else:
            if index_type is None:
                raise Exception(
                    'index_type must be provided if {chunk} is in filename_template'
                )
            if isinstance(chunk, str):
                chunk_str = chunk
            else:
                chunk_str = absorb.ops.format_chunk(chunk, index_type)
            format_params['chunk'] = chunk_str
    return filename_template.format(**format_params)


def get_table_filepaths(
    chunks: typing.Any,
    index_type: absorb.IndexType | None,
    filename_template: str,
    table: str,
    *,
    source: str | None,
    parameters: dict[str, typing.Any],
    warn: bool = True,
) -> list[str]:
    import os

    dir_path = get_table_dir(source=source, table=table, warn=warn)
    paths = []
    for chunk in chunks:
        filename = get_table_filename(
            chunk=chunk,
            index_type=index_type,
            filename_template=filename_template,
            table=table,
            source=source,
            parameters=parameters,
        )
        path = os.path.join(dir_path, filename)
        paths.append(path)
    return paths


def parse_file_path(
    path: str,
    filename_template: str,
    *,
    index_type: absorb.IndexType | None = None,
) -> dict[str, typing.Any]:
    import os

    keys = os.path.splitext(filename_template)[0].split('__')
    values = os.path.splitext(os.path.basename(path))[0].split('__')
    items = {k[1:-1]: v for k, v in zip(keys, values)}
    if index_type is not None and 'chunk' in items:
        items['chunk'] = parse_chunk(items['chunk'], index_type)
    return items


def parse_chunk(as_str: str, index_type: absorb.IndexType) -> typing.Any:
    import datetime

    if index_type == 'hour':
        return datetime.datetime.strptime(as_str, '%Y-%m-%d--%H-%M-%S')
    elif index_type == 'day':
        return datetime.datetime.strptime(as_str, '%Y-%m-%d')
    elif index_type == 'week':
        return datetime.datetime.strptime(as_str, '%Y-%m-%d')
    elif index_type == 'month':
        return datetime.datetime.strptime(as_str, '%Y-%m')
    elif index_type == 'quarter':
        year = int(as_str[:4])
        month = int(as_str[as_str.index('Q') + 1 :])
        return datetime.datetime(year, month, 1)
    elif index_type == 'year':
        return datetime.datetime.strptime(as_str, '%Y')
    elif index_type == 'timestamp':
        import datetime

        return datetime.datetime.strptime(as_str, '%Y-%m-%d')
    else:
        raise NotImplementedError()


def get_dir_size(path: str) -> int:
    import platform
    import subprocess

    system = platform.system()

    if system == 'Linux':
        # Linux has -b flag for bytes
        result = subprocess.run(
            ['du', '-sb', path], capture_output=True, text=True, check=True
        )
        return int(result.stdout.strip().split('\t')[0])

    elif system == 'Darwin':  # macOS
        # macOS outputs in 512-byte blocks
        result = subprocess.run(
            ['du', '-s', path], capture_output=True, text=True, check=True
        )
        blocks = int(result.stdout.strip().split('\t')[0])
        return blocks * 512

    else:
        raise NotImplementedError(
            'Unsupported operating system for get_dir_size'
        )


def format_bytes(bytes_size: int | float) -> str:
    for unit in ['B', 'KB', 'MB', 'GB', 'TB']:
        if bytes_size < 1024.0:
            return f'{bytes_size:.2f} {unit}'
        bytes_size /= 1024.0
    return f'{bytes_size:.2f} PB'


def delete_table_dir(table: absorb.Table, confirm: bool = False) -> None:
    import os
    import shutil

    if not confirm:
        raise absorb.ConfirmError(
            'use confirm=True to delete table and its data files'
        )

    table_dir = table.get_table_dir()
    if os.path.isdir(table_dir):
        shutil.rmtree(table_dir)

    if absorb.ops.get_config()['use_git']:
        absorb.ops.git_remove_and_commit_file(
            absorb.ops.get_table_metadata_path(table),
            repo_root=absorb.ops.get_absorb_root(),
            message='Remove table metadata for ' + table.full_name(),
        )



from __future__ import annotations

import typing

import absorb

if typing.TYPE_CHECKING:
    import polars as pl


class Metric(absorb.Table):
    source = 'fred'
    write_range = 'overwrite_all'
    parameter_types = {'series_id': str}

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Datetime(time_unit='ms'),
            'raw_value': pl.Float64,
            'value': pl.Float64,
            'metric': pl.String,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import polars as pl

        series_id = self.parameters['series_id']
        return get_series(series_id, rename=False).select(
            timestamp=pl.col.date.cast(pl.Datetime('ms')),
            raw_value='raw_value',
            value='value',
            metric=pl.lit(series_id),
        )

    def get_available_range(self) -> absorb.Coverage:
        import datetime

        metadata = get_series_metadata(self.parameters['series_id'])
        return (
            datetime.datetime.strptime(
                metadata['observation_start'], '%Y-%m-%d'
            ),
            datetime.datetime.strptime(metadata['observation_end'], '%Y-%m-%d'),
        )


def get_tables() -> list[type[absorb.Table]]:
    return tables


def get_api_key() -> str:
    import os

    api_key = os.environ['FRED_API_KEY']
    if api_key is None or api_key == '':
        raise Exception('must set FRED_API_KEY')
    return api_key


def get_series_catalog() -> dict[str, tuple[str, str, str]]:
    # (series_id, cadence, name)
    return {
        # monetary
        'M1': ('M1SL', 'month', 'm1'),
        'M1Raw': ('M1NS', 'month', 'm1_raw'),
        'M2': ('M2SL', 'month', 'm2'),
        'M2Raw': ('M2NS', 'month', 'm2_raw'),
        # inflation
        'CPI': ('CPIAUCSL', 'month', 'cpi'),
        'PCEPI': ('PCEPI', 'month', 'pcepi'),
        'PPI': ('PPIACO', 'month', 'ppi'),
        # interest rates
        'InterestRates': ('FEDFUNDS', 'month', 'interest_rates'),
        'BondYield10y': ('DGS10', 'day', 'bond_yield_10y'),
        # labor
        'Unemployment': ('UNRATE', 'month', 'unemployment'),
        'NonfarmEmployment': ('PAYEMS', 'month', 'nonfarm_employment'),
        'LaborForceParticipation': (
            'CIVPART',
            'month',
            'labor_force_participation',
        ),
        # economic output
        'NominalGdp': ('GDP', 'quarter', 'nominal_gdp'),
        'AdjustedGdp': ('GDPC1', 'quarter', 'adjusted_gdp'),
    }


def get_normalized_columns() -> dict[str, tuple[pl.Expr, str]]:
    import polars as pl

    return {
        'Billions of Chained 2017 Dollars': (
            pl.col.raw_value * 1e9,
            '_2017_usd',
        ),
        'Billions of Dollars': (pl.col.raw_value * 1e9, '_usd'),
        'Index 1982-1984=100': (pl.col.raw_value, '_index_1984'),
        'Index 1982=100': (pl.col.raw_value, '_index_1982'),
        'Index 2017=100': (pl.col.raw_value, '_index_2017'),
        'Percent': (pl.col.raw_value, '_percent'),
        'Thousands of Persons': ((pl.col.raw_value * 1000), '_people'),
    }


def _fetch(url: str, params: dict[str, str]) -> dict[str, typing.Any]:
    import requests

    params['api_key'] = get_api_key()
    params['file_type'] = 'json'
    response = requests.get(url, params=params)
    result: dict[str, typing.Any] = response.json()
    return result


def get_series_metadata(name: str) -> dict[str, typing.Any]:
    url = 'https://api.stlouisfed.org/fred/series'
    catalog = get_series_catalog()
    if name in catalog:
        series_id = catalog[name][0]
    else:
        series_id = name
    result = _fetch(url, {'series_id': series_id})
    if len(result['seriess']) == 0:
        raise Exception('series not found')
    elif len(result['seriess']) > 1:
        raise Exception('multiple series found')
    else:
        return result['seriess'][0]  # type: ignore


def get_series(
    name: str,
    *,
    parameters: dict[str, typing.Any] | None = None,
    normalize: bool = True,
    rename: bool = True,
) -> pl.DataFrame:
    import polars as pl

    url = 'https://api.stlouisfed.org/fred/series/observations'

    catalog = get_series_catalog()
    if name in catalog:
        series_id = catalog[name][0]
    else:
        series_id = name
    params = {'series_id': series_id}
    if parameters is not None:
        params.update(parameters)
    result = _fetch(url, params)
    df = pl.DataFrame(result['observations']).select(
        date=pl.col.date.str.to_date(),
        raw_value=pl.col.value.replace({'.': None}).cast(pl.Float64),
    )

    if normalize:
        normalized_columns = get_normalized_columns()
        metadata = get_series_metadata(name)
        if metadata['units'] in normalized_columns:
            column, suffix = normalized_columns[metadata['units']]
            if rename:
                new_name = name + suffix
            else:
                new_name = 'value'
            df = df.with_columns(column.alias(new_name))

    return df


def get_all_tags() -> pl.DataFrame:
    url = 'https://api.stlouisfed.org/fred/tags'

    offset = 0
    tags = []
    print('collecting all tags')
    while True:
        print('fetching', offset)
        result = _fetch(url, params={'offset': str(offset), 'limit': str(1000)})
        tags.extend(result['tags'])
        if len(result['tags']) == 0 or len(tags) >= result['count']:
            break
        else:
            offset += 1000

    return pl.DataFrame(tags)


# create classes
tables: list[type[absorb.Table]] = []
for name, (series_id, cadence, lower_name) in get_series_catalog().items():
    namespace = {'default_parameters': {'series_id': name}}
    cls: Metric = type(name, (Metric,), namespace)  # type: ignore
    cls.description = 'FRED data for ' + name
    cls.url = 'https://fred.stlouisfed.org/series/' + series_id
    cls.index_type = cadence  # type: ignore
    cls.name_template = lower_name
    cls.required_credentials = ['FRED_API_KEY']
    tables.append(cls)  # type: ignore
    globals()[name] = cls



# follow https://docs.sourcify.dev/docs/repository/file-repositories/#download
# 1. get manifest from https://repo-backup.sourcify.dev/
# 2. download all files in manifest
# data is non-tabular, need an alternative to Table



from __future__ import annotations

import typing

import absorb

if typing.TYPE_CHECKING:
    import polars as pl
    from google.cloud import bigquery


class Query(absorb.Table):
    source = 'bigquery'
    description = 'Query data from Google BigQuery'
    url = 'https://cloud.google.com/bigquery?hl=en'
    write_range = 'overwrite_all'
    parameter_types = {'sql': str, 'name': str}
    required_packages = [
        'google-cloud-bigquery >= 3.32.0',
        'google-cloud-bigquery-storage >= 2.31.0',
    ]

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        raise NotImplementedError()

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        return query(self.parameters['sql'])

    def get_available_range(self) -> absorb.Coverage:
        raise NotImplementedError()


def get_client() -> bigquery.Client:
    import google.auth.exceptions
    from google.cloud import bigquery

    try:
        return bigquery.Client()
    except google.auth.exceptions.DefaultCredentialsError as e:
        print(
            'credentials not set up. create a google cloud service account and set the GOOGLE_APPLICATION_CREDENTIALS env variable to the json file of the credentials'  # noqa
        )
        raise e


def query(sql: str) -> pl.DataFrame:
    import pyarrow as pa  # type: ignore
    import pyarrow.compute as pc  # type: ignore

    client = get_client()

    # execute query
    query_job = client.query(sql)

    # retrieve query results as arrow
    results = query_job.result()
    arrow = results.to_arrow()

    # drop nested types
    keep_columns = [
        field.name
        for field in arrow.schema
        if not (
            isinstance(field.type, pa.StructType)
            or isinstance(field.type, pa.ListType)
        )
    ]
    arrow = arrow.select(keep_columns)

    # convert decimal columns to float
    decimal_cols = [
        field.name
        for field in arrow.schema
        if isinstance(field.type, pa.Decimal128Type)
        or isinstance(field.type, pa.Decimal256Type)
    ]
    new_columns = []
    for col_name in arrow.column_names:
        col = arrow[col_name]
        if col_name in decimal_cols:
            casted_col = pc.cast(col, pa.float64())
            new_columns.append(casted_col)
        else:
            new_columns.append(col)
    arrow = pa.Table.from_arrays(new_columns, names=arrow.column_names)

    # convert from arrow to polars
    df: pl.DataFrame = pl.from_arrow(arrow)  # type: ignore

    return df



from __future__ import annotations

from . import common


class BeaconBlocks(common.XatuTable):
    datatype = 'canonical_beacon_block'
    source = 'cannonical_beacon'
    index_type = 'hour'


class BeaconCommittee(common.XatuTable):
    datatype = 'canonical_beacon_committee'
    source = 'cannonical_beacon'
    index_type = 'hour'


class AttesterSlashings(common.XatuTable):
    datatype = 'canonical_beacon_block_attester_slashing'
    source = 'cannonical_beacon'
    index_type = 'hour'


class ProposerSlashings(common.XatuTable):
    datatype = 'canonical_beacon_block_proposer_slashing'
    source = 'cannonical_beacon'
    index_type = 'hour'


class BlockBlsToExecutionChange(common.XatuTable):
    datatype = 'canonical_beacon_block_bls_to_execution_change'
    source = 'cannonical_beacon'
    index_type = 'hour'


class BlockExecutionTransaction(common.XatuTable):
    datatype = 'canonical_beacon_block_execution_transaction'
    source = 'cannonical_beacon'
    index_type = 'hour'


class VoluntaryExits(common.XatuTable):
    datatype = 'canonical_beacon_block_voluntary_exit'
    source = 'cannonical_beacon'
    index_type = 'hour'


class BlockDeposits(common.XatuTable):
    datatype = 'canonical_beacon_block_deposit'
    source = 'cannonical_beacon'
    index_type = 'hour'


class BlockWithdrawals(common.XatuTable):
    datatype = 'canonical_beacon_block_withdrawal'
    source = 'cannonical_beacon'
    index_type = 'hour'


class BlobSidecars(common.XatuTable):
    datatype = 'canonical_beacon_blob_sidecar'
    source = 'cannonical_beacon'
    index_type = 'hour'


class ProposerDuties(common.XatuTable):
    datatype = 'canonical_beacon_proposer_duty'
    source = 'cannonical_beacon'
    index_type = 'hour'


class ElaborateAttestations(common.XatuTable):
    datatype = 'canonical_beacon_elaborated_attestation'
    source = 'cannonical_beacon'
    index_type = 'hour'


class BeaconValidators(common.XatuTable):
    datatype = 'canonical_beacon_validators'
    source = 'cannonical_beacon'
    index_type = 'hour'


class ValidatorsPubkeys(common.XatuTable):
    datatype = 'canonical_beacon_validators_pubkeys'
    source = 'cannonical_beacon'
    index_type = 'hour'



from __future__ import annotations

import argparse
import json
import typing

import rich_argparse

if typing.TYPE_CHECKING:
    Subparsers = argparse._SubParsersAction[argparse.ArgumentParser]


def open_interactive_session(*, variables: dict[str, typing.Any]) -> None:
    print()
    header = (
        'interactive python session started\n\ndata sorted in these variables:'
    )
    for key, value in variables.items():
        header += (
            '\n- \033[1m\033[97m' + key + '\033[0m: ' + type(value).__name__
        )
    try:
        from IPython.terminal.embed import InteractiveShellEmbed

        ipshell = InteractiveShellEmbed(colors='Linux', display_banner=False)  # type: ignore
        print(header)
        ipshell(header=header, local_ns=variables)
    except ImportError:
        import code
        import sys

        class ExitInteract:
            def __call__(self) -> None:
                raise SystemExit

            def __repr__(self) -> str:
                raise SystemExit

        try:
            sys.ps1 = '>>> '
            code.interact(
                banner='\n' + header + '\n',
                local=dict(variables, exit=ExitInteract()),
            )
        except SystemExit:
            pass


def _enter_debugger() -> None:
    """open debugger to most recent exception

    - adapted from https://stackoverflow.com/a/242514
    """
    import sys
    import traceback

    # print stacktrace
    extype, value, tb = sys.exc_info()
    print('[ENTERING DEBUGGER]')

    # print traceback
    try:
        from IPython.core.ultratb import VerboseTB

        vtb = VerboseTB(
            color_scheme='Linux'
        )  # You can also try 'LightBG' or 'Neutral'
        if extype and value and tb:
            formatted_tb = vtb.text(extype, value, tb)
            print(formatted_tb)
        else:
            traceback.print_exc()
    except ImportError:
        traceback.print_exc()
    print()

    # enter debugger
    try:
        import ipdb  # type: ignore
        import types

        tb = typing.cast(types.TracebackType, tb)
        ipdb.post_mortem(tb)

    except ImportError:
        import pdb

        pdb.post_mortem(tb)


class HelpFormatter(rich_argparse.RichHelpFormatter):
    usage_markup = True

    styles = {
        'argparse.prog': 'bold white',
        'argparse.groups': 'bold green',
        'argparse.args': 'bold white',
        'argparse.metavar': 'grey62',
        'argparse.help': 'grey62',
        'argparse.text': 'blue',
        'argparse.syntax': 'blue',
        'argparse.default': 'blue',
    }

    def __init__(self, prog: str) -> None:
        super().__init__(prog, max_help_position=32)

    def _format_args(self, action, default_metavar):  # type: ignore
        get_metavar = self._metavar_formatter(action, default_metavar)
        if action.nargs == argparse.ZERO_OR_MORE:
            return '[%s [%s ...]]' % get_metavar(2)
        elif action.nargs == argparse.ONE_OR_MORE:
            return '%s [...]' % get_metavar(1)
        return super()._format_args(action, default_metavar)

    def format_help(self) -> str:
        lines = [
            line
            for line in super().format_help().split('\n')
            if not line.startswith('  \x1b[1;37m{')
        ]
        return '\n'.join(lines)



from __future__ import annotations

import absorb


def setup_git(track_datasets: list[absorb.Table] | None = None) -> None:
    import os

    absorb_root = absorb.ops.get_absorb_root()

    # initialize repo
    if not git_is_repo_root(absorb_root):
        git_initialize_repo(absorb_root)

    # setup gitignore
    gitignore_path = os.path.join(absorb_root, '.gitignore')
    if not os.path.isfile(gitignore_path):
        default_gitignore = '*.parquet'
        with open(gitignore_path, 'w') as f:
            f.write(default_gitignore)
    if not git_is_file_tracked(gitignore_path, repo_root=absorb_root):
        git_add_and_commit_file(gitignore_path, repo_root=absorb_root)

    # add config file
    config_path = absorb.ops.get_config_path()
    if not git_is_file_tracked(config_path, repo_root=absorb_root):
        git_add_and_commit_file(config_path, repo_root=absorb_root)

    # add metadata of existing tables
    if track_datasets is not None:
        n_added = 0
        for dataset in track_datasets:
            metadata_path = absorb.ops.get_table_metadata_path(dataset)
            if not git_is_file_tracked(metadata_path, repo_root=absorb_root):
                git_add_file(metadata_path, repo_root=absorb_root)
                n_added += 1
        if n_added > 0:
            git_commit(
                message='added ' + str(n_added) + ' table metadata files',
                repo_root=absorb_root,
            )


def git_is_file_tracked(path: str, repo_root: str) -> bool:
    """Check if a file is currently being tracked by git"""
    import subprocess

    try:
        result = subprocess.run(
            ['git', 'ls-files', path],
            cwd=repo_root,
            check=True,
            capture_output=True,
            text=True,
        )
        return bool(result.stdout.strip())
    except subprocess.CalledProcessError:
        return False


def git_initialize_repo(repo_root: str) -> None:
    import os

    if not os.path.isdir(repo_root):
        os.makedirs(repo_root, exist_ok=True)

    cmd = ['git', 'init']
    run_git_command(cmd, repo_root=repo_root)


def git_is_repo_root(repo_root: str) -> bool:
    import os

    return os.path.isdir(os.path.join(repo_root, '.git'))


def git_add_file(path: str, repo_root: str) -> None:
    cmd = ['git', 'add', path]

    run_git_command(cmd, repo_root=repo_root)


def git_add_and_commit_file(
    path: str, repo_root: str, message: str | None = None
) -> None:
    if message is None:
        import os

        message = 'Add ' + os.path.relpath(path, repo_root)

    git_add_file(path=path, repo_root=repo_root)
    git_commit(message=message, repo_root=repo_root)


def git_remove_file(path: str, repo_root: str) -> None:
    cmd = ['git', 'rm', path]

    run_git_command(cmd, repo_root=repo_root)


def git_remove_and_commit_file(
    path: str, repo_root: str, message: str | None = None
) -> None:
    if message is None:
        import os

        message = 'Remove ' + os.path.relpath(path, repo_root)

    git_remove_file(path=path, repo_root=repo_root)
    git_commit(message=message, repo_root=repo_root)


def git_commit(message: str, repo_root: str, *, verbose: bool = False) -> None:
    """Commit staged changes"""
    import subprocess

    # Check if there are staged changes
    try:
        result = subprocess.run(
            ['git', 'diff', '--cached', '--name-only'],
            cwd=repo_root,
            capture_output=True,
            text=True,
            check=True,
        )
        if not result.stdout.strip():
            # No staged changes
            if verbose:
                print('Nothing to commit, working tree clean')
            return
    except subprocess.CalledProcessError as e:
        # If git diff fails, log it but continue with commit attempt
        print(f'Warning: Could not check for staged changes: {e}')

    # Proceed with commit
    try:
        result = subprocess.run(
            ['git', 'commit', '-m', message],
            cwd=repo_root,
            check=True,
            capture_output=True,
            text=True,
        )
        if verbose:
            print(f'Committed: {result.stdout.strip()}')
    except subprocess.CalledProcessError as e:
        error_msg = e.stderr.strip() if e.stderr else str(e)
        print(f'Error committing: {error_msg}')
        raise


def run_git_command(cmd: list[str], repo_root: str) -> str:
    """Run a git command in the specified repository directory"""
    import subprocess

    try:
        result = subprocess.run(
            cmd,
            cwd=repo_root,
            check=True,
            capture_output=True,
            text=True,
        )
        return result.stdout.strip()
    except subprocess.CalledProcessError as e:
        error_msg = e.stderr.strip() if e.stderr else str(e)
        print(f"Error running command '{cmd}': {error_msg}")
        raise
    except Exception as e:
        print(f'An unexpected error occurred: {e}')
        raise



from __future__ import annotations

import typing
import absorb


def _camel_to_snake(name: str) -> str:
    result = []
    for i, char in enumerate(name):
        if char.isupper():
            if i != 0:
                result.append('_')
            result.append(char.lower())
        else:
            result.append(char)
    return ''.join(result)


def _snake_to_camel(name: str) -> str:
    result = []
    capitalize_next = False

    for i, char in enumerate(name):
        if char == '_':
            capitalize_next = True
        elif capitalize_next:
            result.append(char.upper())
            capitalize_next = False
        else:
            result.append(char)

    return result[0].upper() + ''.join(result[1:])


def is_valid_name(name: str) -> bool:
    import re

    if '__' in name or name[0] == '_' or name[-1] == '_':
        return False
    return bool(re.match(r'^[a-zA-Z0-9_]+$', name))


def get_table_name(
    *,
    class_name: str,
    template: str | list[str],
    parameters: dict[str, typing.Any] | None = None,
    allow_generic: bool = False,
) -> str:
    # get template str
    if isinstance(template, list):
        if len(template) == 1:
            template = template[0]
        else:
            for subtemplate in template:
                # get variable names from template (in curly braces)
                template_parameters = _get_template_variables(subtemplate)
                if all(
                    parameters is not None and parameters.get(var) is not None
                    for var in template_parameters
                ):
                    template = subtemplate
                    break
            else:
                raise ValueError(
                    'could not determine template string from list'
                )
    if not isinstance(template, str):
        raise TypeError('template must be a string or a list of strings')

    # assemble the template variables
    if class_name.isupper():
        class_name = class_name.lower()
    else:
        class_name = absorb.ops.names._camel_to_snake(class_name)
    template_vars = {'class_name': class_name}
    if parameters:
        template_vars.update(parameters)

    # whether to allow generic template variables in the output
    if allow_generic:
        result: str = template
        for key, value in template_vars.items():
            result = result.replace(f'{{{key}}}', str(value))
        return result
    else:
        return template.format(**template_vars)


def _get_template_variables(template: str) -> list[str]:
    """
    Extracts variable names from a template string.
    Variables are expected to be in curly braces, e.g., {variable_name}.
    """
    variables = []
    in_variable = False
    current_var: list[str] = []

    for char in template:
        if char == '{':
            in_variable = True
            current_var = []
        elif char == '}':
            if in_variable:
                variables.append(''.join(current_var))
                in_variable = False
        elif in_variable:
            current_var.append(char)

    return variables



from __future__ import annotations

import typing

import absorb
from . import common


if typing.TYPE_CHECKING:
    import polars as pl


# class Fees(absorb.Table):
#     source = 'defillama'
#     write_range = 'overwrite_all'
#     index_type = 'date_range'
#     parameter_types = {
#         'chains': typing.Union[list[str], None],
#         'parameters': typing.Union[list[str], None],
#     }
#     default_parameters = {'chains': None, 'protocols': None}
#     name_template = {
#         'default': 'fees',
#         'custom': {
#             '{chains}': 'chains_fees_{chains}',
#             '{protocols}': 'protocol_fees_{protocols}',
#         },
#     }

#     def get_schema(self) -> dict[str, pl.DataType | pl.DataType | typ ]:
#         import polars as pl

#         return {
#             'timestamp': pl.Datetime('ms'),
#             'chain': pl.String,
#             'protocol': pl.String,
#             'revenue_usd': pl.Int64,
#         }

#     def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
#         import polars as pl

#         if self.parameters['protocols'] is not None:
#             protocols = self.parameters['protocols']
#             if protocols is None:
#                 protocols = _get_fee_protocols()
#             dfs = []
#             print('collecting fees of', len(protocols), 'protocols')
#             for p, protocol in enumerate(protocols, start=1):
#                 print(
#                     '[' + str(p) + ' / ' + str(len(protocols)) + ']', protocol
#                 )
#                 df = get_historical_fees_per_chain_of_protocol(protocol)
#                 dfs.append(df)
#             print('done')
#             return pl.concat(dfs)
#         else:
#             chains = self.parameters['chains']
#             if chains is None:
#                 chains = _get_fee_chains()
#             dfs = []
#             print('collecting fees of', len(chains), 'chains')
#             for c, chain in enumerate(chains, start=1):
#                 print('[' + str(c) + ' / ' + str(len(chains)) + ']', chain)
#                 df = get_historical_fees_per_protocol_of_chain(chain)
#                 dfs.append(df)
#             print('done')
#             return pl.concat(dfs)


class Fees(absorb.Table):
    source = 'defillama'
    description = 'Total fees collected by protocols across all chains'
    url = 'https://defillama.com/'
    write_range = 'overwrite_all'
    index_type = 'day'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Datetime('ms'),
            'revenue_usd': pl.Int64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        return get_historical_fees()


class ChainFees(absorb.Table):
    source = 'defillama'
    description = 'Total fees collected by each protocol and each chain'
    url = 'https://defillama.com/'
    write_range = 'overwrite_all'
    parameter_types = {'chains': (list, type(None))}
    default_parameters = {'chains': None}
    index_type = 'day'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Datetime('ms'),
            'chain': pl.String,
            'protocol': pl.String,
            'revenue_usd': pl.Int64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import polars as pl

        chains = self.parameters['chains']
        if chains is None:
            chains = _get_fee_chains()
        dfs = []
        print('collecting', len(chains), 'chains')
        for c, chain in enumerate(chains, start=1):
            print('[' + str(c) + ' / ' + str(len(chains)) + ']', chain)
            df = get_historical_fees_per_protocol_of_chain(chain)
            df = df.select(
                'timestamp',
                pl.lit(chain).alias('chain'),
                'protocol',
                'revenue_usd',
            )
            dfs.append(df)
        print('done')
        return pl.concat(dfs)


class FeesOfProtocols(absorb.Table):
    source = 'defillama'
    description = 'Total fees collected by each protocol and each chain'
    url = 'https://defillama.com/'
    write_range = 'overwrite_all'
    parameter_types = {'protocols': (list, type(None))}
    default_parameters = {'protocols': None}
    index_type = 'day'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Datetime('ms'),
            'chain': pl.String,
            'protocol': pl.String,
            'revenue_usd': pl.Int64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import polars as pl

        protocols = self.parameters['protocols']
        if protocols is None:
            protocols = _get_fee_protocols()
        dfs = []
        print('collecting', len(protocols), 'protocols')
        for p, protocol in enumerate(protocols, start=1):
            print('[' + str(p) + ' / ' + str(len(protocols)) + ']', protocol)
            df = get_historical_fees_per_chain_of_protocol(protocol)
            dfs.append(df)
        print('done')
        return pl.concat(dfs)


def _get_fee_chains() -> list[str]:
    data = common._fetch('historical_fees')
    chains: list[str] = data['allChains']
    return chains


def _get_fee_protocols() -> list[str]:
    data = common._fetch('historical_fees')
    return list(set(protocol['slug'] for protocol in data['protocols']))


def get_historical_fees() -> pl.DataFrame:
    import polars as pl

    data = common._fetch('historical_fees')

    return (
        pl.DataFrame(
            data['totalDataChart'],
            schema=['timestamp', 'revenue_usd'],
            orient='row',
            strict=False,
        )
        .with_columns((pl.col.timestamp * 1000).cast(pl.Datetime('ms')))
        .sort('timestamp')
    )


def get_historical_fees_per_protocol_of_chain(chain: str) -> pl.DataFrame:
    import polars as pl

    data = common._fetch('historical_fees_per_chain', {'chain': chain})

    return (
        pl.DataFrame(
            [
                [time, chain, protocol, value]
                for time, item in data['totalDataChartBreakdown']
                for protocol, value in item.items()
            ],
            schema=['timestamp', 'chain', 'protocol', 'revenue_usd'],
            orient='row',
        )
        .with_columns((pl.col.timestamp * 1000).cast(pl.Datetime('ms')))
        .sort('timestamp')
    )


def get_historical_fees_per_chain_of_protocol(protocol: str) -> pl.DataFrame:
    import polars as pl

    data = common._fetch('historical_fees_per_protocol', {'protocol': protocol})

    return (
        pl.DataFrame(
            [
                [time, chain, protocol, value]
                for time, item in data['totalDataChartBreakdown']
                for chain, subitem in item.items()
                for _, value in subitem.items()
            ],
            schema=['timestamp', 'chain', 'protocol', 'revenue_usd'],
            orient='row',
        )
        .with_columns((pl.col.timestamp * 1000).cast(pl.Datetime('ms')))
        .sort('timestamp')
    )



from __future__ import annotations

import typing

import absorb

if typing.TYPE_CHECKING:
    import polars as pl


endpoints = {
    'coin_list': 'https://api.coingecko.com/api/v3/coins/list',
    'category_list': 'https://api.coingecko.com/api/v3/coins/categories/list',
    'current_coin_prices': 'https://api.coingecko.com/api/v3/coins/markets',
    'historical_coin_prices': 'https://api.coingecko.com/api/v3/coins/{id}/market_chart',
}


class CoinMetrics(absorb.Table):
    source = 'coingecko'
    description = 'Price, market cap, and volume data for coins'
    url = 'https://coingecko.com/'
    write_range = 'overwrite_all'
    index_type = 'day'
    parameter_types = {'top_n': int}
    default_parameters = {'top_n': None}
    name_template = [
        'coin_metrics_top_{top_n}',
        'coin_metrics',
    ]

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Datetime(time_unit='ms'),
            'coin': pl.String,
            'price': pl.Float64,
            'market_cap_usd': pl.Float64,
            'volume_usd': pl.Float64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        top_n = self.parameters['top_n']
        if top_n is None:
            top_n = 1000
        return get_historical_coin_metrics(top_n)

    def get_available_range(self) -> absorb.Coverage:
        import datetime

        now = datetime.datetime.now()
        now = datetime.datetime(year=now.year, month=now.month, day=now.day)
        return (now - datetime.timedelta(days=364), now)


class Categories(absorb.Table):
    source = 'coingecko'
    description = 'Categorizations of coins'
    url = 'https://coingecko.com/'
    write_range = 'overwrite_all'
    index_type = 'id'
    parameter_types = {'categories': (list, type(None))}
    default_parameters = {'categories': None}
    name_template = [
        'coin_metrics_{categories}',
        'coin_metrics',
    ]

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {'coin': pl.String, 'category': pl.String}

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        return get_current_coin_categories(
            categories=self.parameters['categories']
        )

    def get_available_range(self) -> absorb.Coverage:
        import datetime

        now = datetime.datetime.now()
        now = datetime.datetime(year=now.year, month=now.month, day=now.day)
        return (now, now)


class CategoryMetrics(absorb.Table):
    source = 'coingecko'
    description = 'Aggregated metrics for each category of coins'
    url = 'https://coingecko.com/'
    write_range = 'overwrite_all'
    index_type = 'day'
    parameter_types = {'categories': (list, type(None))}
    default_parameters = {'categories': None}
    dependencies = [CoinMetrics, Categories]
    name_template = [
        'category_metrics_{categories}',
        'category_metrics',
    ]

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Datetime(time_unit='ms'),
            'category': pl.String,
            'market_cap_usd': pl.Float64,
            'volume_usd': pl.Float64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        coin_metrics = absorb.load('coingecko.coin_metrics')
        coin_categories = absorb.load('coingecko.categories')
        return get_historical_category_metrics(
            coin_metrics=coin_metrics, coin_categories=coin_categories
        )

    def get_available_range(self) -> absorb.Coverage:
        import datetime

        now = datetime.datetime.now()
        now = datetime.datetime(year=now.year, month=now.month, day=now.day)
        return (now, now)


def _fetch(
    datatype: str,
    *,
    url_params: dict[str, typing.Any] | None = None,
    params: dict[str, typing.Any] | None = None,
    api_key: str | None = None,
) -> typing.Any:
    import time
    import requests

    headers = {'accept': 'application/json'}
    if api_key is None:
        api_key = get_coinbase_api_key()
    if api_key is not None:
        headers['x-cg-demo-api-key'] = api_key

    if url_params is None:
        url_params = {}
    url = endpoints[datatype].format(**url_params)
    # response = session.get(url, headers=headers, params=params)

    n_attempts = 5
    for i in range(n_attempts):
        response = requests.get(url, headers=headers, params=params)
        if response.status_code == 429:
            time.sleep(60 * (i + 1))
            continue
        response.raise_for_status()
        return response.json()
    raise Exception(
        'Failed to fetch data from CoinGecko API after '
        + str(n_attempts)
        + ' attempts.'
    )


def get_coinbase_api_key() -> str | None:
    import os

    return os.environ.get('COINGECKO_API_KEY')


#
# # dataframe fetching
#


def get_coin_list() -> pl.DataFrame:
    import polars as pl

    result = _fetch('coin_list')
    return pl.DataFrame(result)


def get_category_list(top_n: int | None = None) -> pl.DataFrame:
    import polars as pl

    return pl.DataFrame(_fetch('category_list'))


def get_current_coin_prices(
    top_n: int = 1000,
    *,
    category: str | None = None,
    include_price_changes: bool = False,
) -> pl.DataFrame:
    import time
    import math
    import polars as pl

    n_pages = math.ceil(top_n / 250)
    results = []
    for page in range(1, n_pages + 1):
        params = {'vs_currency': 'usd', 'per_page': 250, 'page': page}
        if category is not None:
            params['category'] = category
        if include_price_changes:
            params['price_change_percentage'] = '7d,14d,30d,200d,1y'
        result = _fetch('current_coin_prices', params=params)
        results.extend(result)
        time.sleep(6)
        if len(result) == 0:
            break
    schema = {
        'id': pl.String,
        'symbol': pl.String,
        'name': pl.String,
        'image': pl.String,
        'current_price': pl.Float64,
        'market_cap': pl.Float64,
        'market_cap_rank': pl.Int64,
        'fully_diluted_valuation': pl.Float64,
        'total_volume': pl.Float64,
        'high_24h': pl.Float64,
        'low_24h': pl.Float64,
        'price_change_24h': pl.Float64,
        'price_change_percentage_24h': pl.Float64,
        'market_cap_change_24h': pl.Float64,
        'market_cap_change_percentage_24h': pl.Float64,
        'circulating_supply': pl.Float64,
        'total_supply': pl.Float64,
        'max_supply': pl.Float64,
        'ath': pl.Float64,
        'ath_change_percentage': pl.Float64,
        'ath_date': pl.String,
        'atl': pl.Float64,
        'atl_change_percentage': pl.Float64,
        'atl_date': pl.String,
        # 'roi': pl.Struct({'times': Float64, 'currency': String, 'percentage': Float64}),
        'last_updated': pl.String,
    }
    return pl.DataFrame(results, schema=schema, infer_schema_length=top_n)[
        :top_n
    ]


def get_historical_coin_metrics(
    coins: pl.Series | list[str] | int | None = None,
) -> pl.DataFrame:
    import time
    import polars as pl

    if coins is None:
        coins = get_current_coin_prices()['id']
        time.sleep(5)
    elif isinstance(coins, int):
        coins = get_current_coin_prices(coins)['id']
        time.sleep(5)

    print('getting historical data for', len(coins), 'coins')
    dfs = []
    for t, coin in enumerate(coins, start=1):
        print('[' + str(t) + ' / ' + str(len(coins)) + '] ' + coin)

        # get data
        params = {'vs_currency': 'usd', 'days': 365, 'interval': 'daily'}
        result = _fetch(
            'historical_coin_prices', url_params={'id': coin}, params=params
        )

        # parse into dataframes
        schema: dict[str, pl.DataType | type[pl.DataType]]
        schema = {'timestamp': pl.Datetime('ms'), 'price': pl.Float64}
        prices = pl.DataFrame(result['prices'], schema=schema, orient='row')
        schema = {'timestamp': pl.Datetime('ms'), 'market_cap_usd': pl.Float64}
        cap = pl.DataFrame(result['market_caps'], schema=schema, orient='row')
        schema = {'timestamp': pl.Datetime('ms'), 'volume_usd': pl.Float64}
        vol = pl.DataFrame(result['total_volumes'], schema=schema, orient='row')
        df = prices.join(cap, on='timestamp').join(vol, on='timestamp')
        df = df.insert_column(1, pl.lit(coin).alias('coin'))
        dfs.append(df)

        time.sleep(6)

    return pl.concat(dfs)


def get_current_coin_categories(
    categories: list[str] | pl.Series | None = None,
) -> pl.DataFrame:
    import time
    import polars as pl

    if categories is None:
        categories = get_category_list()['category_id']
        time.sleep(10)

    print('getting tokens for', len(categories), 'categories')
    dfs = []
    for c, category in enumerate(categories, start=1):
        print(str(c) + '. ' + category)
        result = get_current_coin_prices(category=category)
        if len(result) > 0:
            df = result.select(coin='id', category=pl.lit(category))
            dfs.append(df)
        time.sleep(30)
    return pl.concat(dfs)


def get_historical_category_metrics(
    *,
    coin_metrics: pl.DataFrame | None = None,
    coin_categories: pl.DataFrame | None = None,
) -> pl.DataFrame:
    import polars as pl

    if coin_metrics is None:
        coin_metrics = get_historical_coin_metrics()
    if coin_categories is None:
        coin_categories = get_current_coin_categories()
    return (
        coin_metrics.join(coin_categories, on='coin', how='left')
        .group_by('timestamp', 'category')
        .agg(pl.sum('market_cap_usd'), pl.sum('volume_usd'))
        .sort('timestamp', 'category')
    )


def _convert_current_prices_to_changes(prices: pl.DataFrame) -> pl.DataFrame:
    import datetime

    intervals = ['7d', '14d', '30d', '200d', '1y']

    volume_columns = {
        ('value_' + interval): pl.col.current_price
        / (
            1
            + pl.col('price_change_percentage_' + interval + '_in_currency')
            / 100
        )
        for interval in intervals
    }
    abs_delta_columns = {
        'abs_delta_' + interval: pl.col.value_now - pl.col('value_' + interval)
        for interval in intervals
    }
    rel_delta_columns = {
        'rel_delta_' + interval: (
            pl.col.value_now - pl.col('value_' + interval)
        )
        / pl.col('value_' + interval)
        for interval in intervals
    }
    t_columns = {
        't_' + interval: pl.col.t_now.dt.offset_by('-' + interval)
        for interval in intervals
    }

    return (
        prices.select(
            token='id',
            value_now='current_price',
            **volume_columns,
        )
        .with_columns(
            **abs_delta_columns,
            **rel_delta_columns,
            t_now=pl.lit(datetime.datetime.now()),
        )
        .with_columns(
            **t_columns,
        )
    )



from __future__ import annotations

import typing

import absorb

if typing.TYPE_CHECKING:
    import polars as pl


class Query(absorb.Table):
    source = 'snowflake'
    write_range = 'overwrite_all'
    description = 'Snowflake SQL query'
    url = 'https://www.snowflake.com/en/'
    # parameter_types = {'name': str, 'sql': str}
    # parameter_types = {'name': str}
    required_packages = ['paradigm_garlic >= 0.1.2']
    sql: str
    # name_template = 'snowflake_query_{name}'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import garlic

        sql = (
            'WITH cte AS ('
            + self.sql.rstrip(';')
            + '), SELECT * FROM cte LIMIT 0'
        )
        result = garlic.query(sql)
        return dict(result.schema)

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import garlic

        if not hasattr(self, 'sql'):
            raise ValueError('SQL query must be defined in the subclass.')

        return garlic.query(self.sql)



from __future__ import annotations

import typing

import absorb
from . import paths

if typing.TYPE_CHECKING:
    import typing_extensions


def get_default_config() -> absorb.Config:
    import os
    import subprocess

    output = subprocess.check_output(['which', 'git'], text=True)
    use_git = os.path.isfile(output.strip())

    return {
        'version': absorb.__version__,
        'tracked_tables': [],
        'use_git': use_git,
        'default_bucket': {
            'provider': None,
            'bucket_name': None,
            'rclone_remote': None,
            'path_prefix': None,
        },
    }


def get_config() -> absorb.Config:
    import json

    try:
        with open(paths.get_config_path(), 'r') as f:
            config = json.load(f)
    except FileNotFoundError:
        return get_default_config()

    default_config = get_default_config()
    default_config.update(config)
    config = default_config

    if validate_config(config):
        return config
    else:
        raise Exception('invalid config format')


def write_config(
    config: absorb.Config, commit_message: str | None = None
) -> None:
    import json
    import os

    default_config = get_default_config()
    default_config.update(config)
    config = default_config

    if not validate_config(config):
        raise Exception('invalid config format')

    path = paths.get_config_path()
    os.makedirs(os.path.dirname(path), exist_ok=True)

    # load old config for change detection
    if os.path.isfile(path):
        with open(path, 'r') as f:
            old_config = json.load(f)
    else:
        old_config = None

    # write config
    with open(path, 'w') as f:
        json.dump(config, f)

    # version control
    if config['use_git']:
        if json.dumps(config, sort_keys=True) != json.dumps(
            old_config, sort_keys=True
        ):
            absorb.ops.git_add_and_commit_file(
                paths.get_config_path(),
                repo_root=paths.get_absorb_root(),
                message=commit_message,
            )


def validate_config(
    config: typing.Any,
) -> typing_extensions.TypeGuard[absorb.Config]:
    return (
        isinstance(config, dict)
        and {'tracked_tables'}.issubset(set(config.keys()))
        and isinstance(config['tracked_tables'], list)
    )


#
# # specific accessors
#


def get_tracked_tables() -> list[absorb.TableDict]:
    return get_config()['tracked_tables']


def start_tracking_tables(
    tables: typing.Sequence[absorb.TableReference],
) -> None:
    import json

    config = get_config()
    tracked_tables = {
        json.dumps(table, sort_keys=True) for table in config['tracked_tables']
    }
    for table in tables:
        instance = absorb.Table.instantiate(table)
        table_dict = instance.create_table_dict()
        as_str = json.dumps(table_dict, sort_keys=True)

        # check for name collisions
        for tracked_table in config['tracked_tables']:
            if table_dict['table_name'] == tracked_table[
                'table_name'
            ] and as_str != json.dumps(tracked_table, sort_keys=True):
                raise Exception(
                    'name collision, cannot add new tracked table: '
                    + str(table_dict['table_name'])
                )

        # add to tracked tables
        if as_str not in tracked_tables:
            config['tracked_tables'].append(table_dict)
            tracked_tables.add(as_str)

    names = ', '.join(
        absorb.Table.instantiate(table).full_name() for name in tables
    )
    message = 'Start tracking ' + str(len(tables)) + ' tables: ' + names
    write_config(config, message)


def stop_tracking_tables(
    tables: typing.Sequence[absorb.TableReference],
) -> None:
    import json

    drop_these = set()
    for table in tables:
        table_dict = absorb.Table.instantiate(table).create_table_dict()
        table_str = json.dumps(table_dict, sort_keys=True)
        drop_these.add(table_str)

    config = get_config()
    config['tracked_tables'] = [
        table
        for table in config['tracked_tables']
        if json.dumps(table, sort_keys=True) not in drop_these
    ]

    names = ', '.join(
        absorb.Table.instantiate(table).full_name() for table in tables
    )
    message = 'Stop tracking ' + str(len(tables)) + ' tables: ' + names
    write_config(config, message)


def enable_git_tracking() -> None:
    config = get_config()
    config['use_git'] = True
    write_config(config, 'Enable git tracking')


def disable_git_tracking() -> None:
    config = get_config()
    config['use_git'] = False
    write_config(config, 'Disable git tracking')


#
# # environment
#


def is_package_installed(package: str) -> bool:
    """
    Check if a package is installed, optionally with version constraints.

    Args:
        package: Package name with optional version specifier
                 e.g., 'numpy', 'numpy>=1.20', 'numpy==1.21.0'

    Returns:
        bool: True if package is installed and meets version requirements
    """
    import importlib.metadata
    import re

    # Parse package string to extract name and version specifier
    match = re.match(r'^([a-zA-Z0-9_-]+)\s*(.*)$', package.strip())
    if not match:
        return False

    package_name = match.group(1)
    version_spec = match.group(2).strip()

    try:
        # Get installed version
        installed_version = importlib.metadata.version(package_name)
    except importlib.metadata.PackageNotFoundError:
        return False

    # If no version specifier, just check if installed
    if not version_spec:
        return True

    # Parse version specifier
    spec_match = re.match(r'^(==|!=|<=|>=|<|>|~=)\s*(.+)$', version_spec)
    if not spec_match:
        return False

    operator = spec_match.group(1)
    required_version = spec_match.group(2).strip()

    # Compare versions
    return compare_versions(installed_version, operator, required_version)


def compare_versions(installed: str, operator: str, required: str) -> bool:
    """Compare version strings."""
    import re

    # Convert version strings to tuples of integers for comparison
    def version_tuple(v: str) -> tuple[int, ...]:
        return tuple(int(x) for x in re.findall(r'\d+', v))

    installed_tuple = version_tuple(installed)
    required_tuple = version_tuple(required)

    if operator == '==':
        return installed_tuple == required_tuple
    elif operator == '!=':
        return installed_tuple != required_tuple
    elif operator == '>=':
        return installed_tuple >= required_tuple
    elif operator == '<=':
        return installed_tuple <= required_tuple
    elif operator == '>':
        return installed_tuple > required_tuple
    elif operator == '<':
        return installed_tuple < required_tuple
    elif operator == '~=':
        # Compatible release: version should be >= X.Y and < X+1.0
        if len(required_tuple) < 2:
            return False
        return (
            installed_tuple[: len(required_tuple) - 1] == required_tuple[:-1]
            and installed_tuple >= required_tuple
        )

    return False


def set_default_rclone_remote(rclone_remote: str) -> None:
    config = get_config()
    config['default_bucket']['rclone_remote'] = rclone_remote
    write_config(config, 'Set default rclone remote to ' + rclone_remote)


def set_default_bucket(bucket: str) -> None:
    config = get_config()
    config['default_bucket']['bucket_name'] = bucket
    write_config(config, 'Set default bucket to ' + bucket)


def set_default_provider(provider: str) -> None:
    config = get_config()
    config['default_bucket']['provider'] = provider
    write_config(config, 'Set default provider to ' + provider)


def set_default_path_prefix(path_prefix: str) -> None:
    config = get_config()
    config['default_bucket']['path_prefix'] = path_prefix
    write_config(config, 'Set default path prefix to ' + path_prefix)


def clear_default_rclone_remote() -> None:
    config = get_config()
    config['default_bucket']['rclone_remote'] = None
    write_config(config, 'Cleared default rclone remote')


def clear_default_bucket() -> None:
    config = get_config()
    config['default_bucket']['bucket_name'] = None
    write_config(config, 'Cleared default bucket')


def clear_default_provider() -> None:
    config = get_config()
    config['default_bucket']['provider'] = None
    write_config(config, 'Cleared default provider')


def clear_default_path_prefix() -> None:
    config = get_config()
    config['default_bucket']['path_prefix'] = None
    write_config(config, 'Cleared default path prefix')



from __future__ import annotations

from . import common


class BeaconAttestations(common.XatuTable):
    datatype = 'libp2p_gossipsub_beacon_attestation'
    source = 'libp2p'
    index_type = 'hour'


class BeaconBlocks(common.XatuTable):
    datatype = 'libp2p_gossipsub_beacon_block'
    source = 'libp2p'
    index_type = 'hour'


class BlobSidecars(common.XatuTable):
    datatype = 'libp2p_gossipsub_blob_sidecar'
    source = 'libp2p'
    index_type = 'hour'



from __future__ import annotations

import functools
import typing
import absorb

if typing.TYPE_CHECKING:
    import datetime
    import polars as pl


root_url = 'https://data.ethpandaops.io/xatu'

url_templates = {
    'per_day': root_url
    + '{network}/databases/default/{datatype}/{year}/{month}/{day}.parquet',
    'per_hour': root_url
    + '{network}/databases/default/{datatype}/{year}/{month}/{day}/{hour}.parquet',
}


class XatuTable(absorb.Table):
    source: str
    description = 'Xatu data from EthPandaOps'
    url = 'https://ethpandaops.io/data/xatu/'
    datatype: str
    per: typing.Literal['day', 'hour']
    parameter_types = {'network': str}
    name_template = '{class_name}_{network}'
    required_packages = ['yaml >= 5.4.1']

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        return _fetch(
            datatype=self.datatype,
            network=self.parameters['network'],
            timestamp=chunk,  # type: ignore
            per=self.per,
        )

    def get_available_range(self) -> absorb.Coverage:
        return get_table_range(
            table=self.datatype, network=self.parameters['network']
        )


def _fetch(
    datatype: str, network: str, timestamp: datetime.datetime, per: str
) -> pl.DataFrame:
    url_template = url_templates['per_' + per]
    url = url_template.format(
        network=network,
        datatpye=datatype,
        year=timestamp.year,
        month=timestamp.month,
        day=timestamp.day,
        hour=timestamp.hour,
    )
    return absorb.ops.download_parquet_to_dataframe(url)


@functools.lru_cache()
def get_manifest() -> dict[str, typing.Any]:
    import requests
    import yaml  # type: ignore

    url = 'https://raw.githubusercontent.com/ethpandaops/xatu-data/refs/heads/master/config.yaml'
    response = requests.get(url)
    result: dict[str, typing.Any] = yaml.safe_load(response.text)
    return result


def get_table_range(
    table: str, network: str
) -> tuple[int, int] | tuple[datetime.datetime, datetime.datetime]:
    manifest = get_manifest()
    for item in manifest['tables']:
        if item['name'] == table:
            break
    else:
        raise Exception('could not find manifest for table')
    network_range = item['networks'][network]
    start = network_range['from']
    end = network_range['to']
    if item['partitioning']['type'] == 'integer':
        return int(start), int(end)
    elif item['partitioning']['type'] == 'datetime':
        import datetime

        return (
            datetime.datetime.strptime(start, '%Y-%m-%d'),
            datetime.datetime.strptime(end, '%Y-%m-%d'),
        )
    else:
        raise Exception()



from __future__ import annotations

import typing

from . import table_collect
from . import table_create

if typing.TYPE_CHECKING:
    T = typing.TypeVar('T')


class Table(
    table_collect.TableCollect,
    table_create.TableCreate,
):
    pass



from __future__ import annotations

import typing
from . import table_io
import absorb

if typing.TYPE_CHECKING:
    T = typing.TypeVar('T')

    import datetime


class TableCoverage(table_io.TableIO):
    def get_available_range(self) -> absorb.Coverage | None:
        if self.write_range == 'overwrite_all':
            return None
        else:
            raise NotImplementedError(
                'get_available_range() not implemented for '
                + self.source
                + '.'
                + str(type(self).__name__)
            )

    def get_collected_range(self) -> absorb.Coverage | None:
        import os
        import glob

        dir_path = self.get_table_dir()
        if not os.path.isdir(dir_path):
            return None

        glob_str = self.get_glob()
        if self.write_range == 'overwrite_all':
            files = sorted(glob.glob(glob_str))
            if len(files) == 0:
                return None
            elif len(files) == 1:
                import polars as pl

                # for now: only handle timestamp ranges if timestamp present
                schema = self.scan().collect_schema()
                if 'timestamp' in schema.names():
                    df = (
                        self.scan()
                        .select(
                            min_timestamp=pl.col.timestamp.min(),
                            max_timestamp=pl.col.timestamp.max(),
                        )
                        .collect()
                    )
                    return (df['min_timestamp'][0], df['max_timestamp'][0])

                else:
                    return None

                # parsed: dict[str, typing.Any] = self.parse_file_path(files[0])
                # if 'chunk' in parsed:
                #     return [parsed['chunk']]
                # else:
                #     raise Exception('chunk not in name template')
            else:
                raise Exception('too many files')
        elif self.is_range_sortable():
            files = sorted(glob.glob(glob_str))
            if len(files) == 0:
                return None
            start = self.parse_file_path(files[0])['chunk']
            end = self.parse_file_path(files[-1])['chunk']
            return (start, end)
        else:
            raise Exception()

    def get_missing_ranges(self) -> absorb.Coverage | None:
        collected_range = self.get_collected_range()
        available_range = self.get_available_range()
        if available_range is None:
            return None
        if collected_range is None:
            return [available_range]
        else:
            index_type = self.index_type
            if index_type is None:
                raise Exception(
                    'ranges computations require index_type to be set'
                )
            return absorb.ops.get_range_diff(
                subtract_this=collected_range,
                from_this=available_range,
                index_type=index_type,
            )

    @classmethod
    def is_range_sortable(cls) -> bool:
        return cls.index_type is not None

    def ready_for_update(self) -> bool:
        """used for periodically updating datasets that have no get_available_range()"""
        import datetime
        import tooltime

        # ensure valid for dataset
        if not absorb.ops.index_is_temporal(self.index_type):
            raise Exception(
                'ready_for_update() can only be called if index is temporal'
            )
        if not isinstance(self.index_type, str):
            raise Exception(
                'ready_for_update() can only be called if index_type is a string'
            )

        # get last update time
        last_update_time = self.get_max_collected_timestamp()
        if last_update_time is None:
            return True

        # get min latency
        min_latency_seconds: float | int
        if self.update_latency is None:
            min_latency_seconds = tooltime.timelength_to_seconds(
                '1 ' + self.index_type
            )
        elif isinstance(self.update_latency, str):
            min_latency_seconds = tooltime.timelength_to_seconds(
                self.update_latency
            )
        elif isinstance(self.update_latency, float):
            min_latency_seconds = (
                self.update_latency
                * tooltime.timelength_to_seconds('1 ' + self.index_type)
            )
        else:
            raise Exception('invalid format for update_latency')
        min_latency = datetime.timedelta(seconds=min_latency_seconds)

        # return whether now is past the last update time + min latency
        return datetime.datetime.now() > last_update_time + min_latency

    def get_min_collected_timestamp(self) -> datetime.datetime | None:
        import polars as pl

        return self.scan().select(pl.col.timestamp.min()).collect().item()  # type: ignore

    def get_max_collected_timestamp(self) -> datetime.datetime | None:
        import polars as pl

        return self.scan().select(pl.col.timestamp.max()).collect().item()  # type: ignore

    def get_collected_timestamp_range(
        self,
    ) -> tuple[datetime.datetime | None, datetime.datetime | None]:
        import polars as pl

        return (
            self.scan()
            .select(pl.col.timestamp.min(), pl.col.timestamp.max())
            .collect()
            .row(0)
        )



from __future__ import annotations

import typing

import absorb

from . import table_base


class TableNames(table_base.TableBase):
    @classmethod
    def full_class_name(cls) -> str:
        return cls.__module__ + '.' + cls.__qualname__

    @classmethod
    def name_classmethod(
        cls,
        allow_generic: bool = False,
        parameters: dict[str, absorb.JSONValue] | None = None,
    ) -> str:
        # build class parameters
        if parameters is not None:
            parameters = dict(cls.default_parameters, **parameters)
        else:
            parameters = cls.default_parameters
        return absorb.ops.get_table_name(
            class_name=cls.__name__,
            template=cls.name_template,
            parameters=parameters,
            allow_generic=allow_generic,
        )

    def name(self) -> str:
        return absorb.ops.get_table_name(
            class_name=type(self).__name__,
            template=self.name_template,
            parameters=self.parameters,
        )

    def full_name_classmethod(
        cls,
        allow_generic: bool = False,
        parameters: dict[str, absorb.JSONValue] | None = None,
    ) -> str:
        return (
            cls.source
            + '.'
            + cls.name_classmethod(
                allow_generic=allow_generic, parameters=parameters
            )
        )

    def full_name(self) -> str:
        return self.source + '.' + self.name()

    @classmethod
    def parse_name_parameters(cls, table_name: str) -> dict[str, typing.Any]:
        name_templates = (
            cls.name_template
            if isinstance(cls.name_template, list)
            else [cls.name_template]
        )
        for template in name_templates:
            template = template.replace(
                '{class_name}', absorb.ops.names._camel_to_snake(cls.__name__)
            )
            try:
                raw = absorb.ops.parse_string_from_template(
                    template, table_name
                )
            except absorb.NameParseError:
                continue
            return absorb.ops.convert_raw_parameter_types(
                raw, cls.parameter_types
            )
        else:
            raise absorb.NameParseError('Could not parse ' + table_name)



from __future__ import annotations

import typing

import absorb

if typing.TYPE_CHECKING:
    import polars as pl


def scan(
    table: absorb.TableReference,
    *,
    bucket: bool | absorb.Bucket = False,
    scan_kwargs: dict[str, typing.Any] | None = None,
) -> pl.LazyFrame:
    if bucket:
        if isinstance(bucket, bool):
            bucket = absorb.ops.get_default_bucket()
        return absorb.ops.scan_bucket(
            table=table, bucket=bucket, scan_kwargs=scan_kwargs
        )
    else:
        table = absorb.Table.instantiate(table)
        return table.scan(scan_kwargs=scan_kwargs)


def load(
    table: absorb.TableReference,
    *,
    bucket: bool | absorb.Bucket = False,
    scan_kwargs: dict[str, typing.Any] | None = None,
) -> pl.DataFrame:
    """kwargs are passed to scan()"""
    table = absorb.Table.instantiate(table)
    return table.load(scan_kwargs=scan_kwargs)


def write_file(*, df: pl.DataFrame, path: str) -> None:
    import os
    import shutil

    dirname = os.path.dirname(path)
    if dirname != '':
        os.makedirs(dirname, exist_ok=True)

    tmp_path = path + '_tmp'
    if path.endswith('.parquet'):
        df.write_parquet(tmp_path)
    elif path.endswith('.csv'):
        df.write_csv(tmp_path)
    else:
        raise Exception('invalid file extension')
    shutil.move(tmp_path, path)



"""
file organization style
- defillama__yields_per_pool__uniswap_v2__2025-01-01--01-01-02.000.parquet
- defillama__fees_per_chain__solana__2025-01-01--01-01-02.parquet
^ timestamp in filename is time that file was collected or last timesatmp in file?
"""

from __future__ import annotations

import typing


default_root = 'https://api.llama.fi'
stablecoin_root = 'https://stablecoins.llama.fi'
yield_root = 'https://yields.llama.fi'


endpoints = {
    # tvl
    'current_tvls': default_root + '/protocols',
    'historical_tvl_of_protocol': default_root + '/protocol/{protocol}',
    'historical_tvl': default_root + '/v2/historicalChainTvl',
    'historical_tvl_of_chain': default_root + '/v2/historicalChainTvl/{chain}',
    'current_tvl_of_protocol': default_root + '/tvl/{protocol}',
    'current_tvl_per_chain': default_root + '/v2/chains',
    # stablecoins
    'current_stablecoins': stablecoin_root + '/stablecoins',
    'historical_total_stablecoins': stablecoin_root + '/stablecoincharts/all',
    'historical_stablecoins_of_chain': stablecoin_root
    + '/stablecoincharts/{chain}',
    'historical_stablecoins_of_token': stablecoin_root + '/stablecoin/{token}',
    'current_stablecoins_per_chain': stablecoin_root + '/stablecoinchains',
    'historical_stablecoin_prices': stablecoin_root + '/stablecoinprices',
    # yields
    'current_yields': yield_root + '/pools',
    'historical_yields_per_pool': yield_root + '/chart/{pool}',
    # volumes
    'historical_dex_volume': default_root + '/overview/dexs',
    'historical_dex_volume_of_chain': default_root + '/overview/dexs/{chain}',
    'historical_dex_volume_of_protocol': default_root
    + '/summary/dexs/{protocol}',
    'historical_options_volume': default_root + '/overview/options',
    'historical_options_volume_of_chain': default_root
    + '/overview/options/{chain}',
    'historical_options_volume_of_protocol': default_root
    + '/summary/options/{protocol}',
    # fees and revenue
    'historical_fees': default_root + '/overview/fees',
    'historical_fees_per_chain': default_root + '/overview/fees/{chain}',
    'historical_fees_per_protocol': default_root + '/summary/fees/{protocol}',
}


def _fetch(
    endpoint: str, parameters: dict[str, str] | None = None
) -> typing.Any:
    import requests

    url = _get_url(endpoint, parameters)
    response = requests.get(url, timeout=(5, 60))
    response.raise_for_status()
    return response.json()


def _get_url(
    endpoint: str, parameters: dict[str, str] | None = None
) -> typing.Any:
    url = endpoints[endpoint]
    if parameters is not None:
        url = url.format(**parameters)
    return url



from __future__ import annotations

from . import common


class BidTraces(common.XatuTable):
    datatype = 'mev_relay_bid_trace'
    source = 'mev_relay'
    index_type = 'hour'


class ProposerPayloadDeliveries(common.XatuTable):
    datatype = 'mev_relay_proposer_payload_delivered'
    source = 'mev_relay'
    index_type = 'hour'


class ValidatorRegistrations(common.XatuTable):
    datatype = 'mev_relay_validator_registration'
    source = 'mev_relay'
    index_type = 'hour'



from __future__ import annotations

import typing

import absorb


def parse_table_str(
    ref: str,
    *,
    parameters: dict[str, typing.Any] | None = None,
    raw_parameters: dict[str, str] | None = None,
    use_all_parameters: bool = True,
    allow_generic: bool = False,
) -> tuple[type[absorb.Table], dict[str, typing.Any]]:
    """
    Parses a table string of the form 'source.table_name' and returns the class
    and parameters.
    """
    source, table_name = ref.split('.')

    for cls in absorb.ops.get_source_table_classes(source):
        try:
            name_parameters = cls.parse_name_parameters(table_name)
            for key, value in name_parameters.items():
                if '{' + key + '}' == value and not allow_generic:
                    raise absorb.NameParseError(
                        f'Generic parameter {key} not allowed in {ref}'
                    )
            break
        except absorb.NameParseError:
            continue
    else:
        raise Exception('Could not find table class for: ' + ref)

    # parse input parameters
    if parameters is None:
        parameters = {}
    if raw_parameters is not None:
        parameters.update(
            convert_raw_parameter_types(raw_parameters, cls.parameter_types)
        )

    # merge name parameters into input parameters
    parameters = dict(parameters, **name_parameters)

    # only use subset relevant to Table class
    for key, value in list(parameters.items()):
        if key not in cls.parameter_types:
            if not use_all_parameters:
                del parameters[key]
            else:
                raise Exception(
                    'Invalid parameter ' + key + ' for ' + cls.full_class_name()
                )

    return cls, parameters


def parse_string_from_template(template: str, string: str) -> dict[str, str]:
    import re

    # Find all placeholders in the template
    placeholder_pattern = r'\{([^}]+)\}'
    placeholders = re.findall(placeholder_pattern, template)

    # Escape special regex characters in the template, but preserve placeholders
    regex_pattern = re.escape(template)

    # Replace escaped placeholders with capture groups
    # Use non-greedy matching by default to handle multiple placeholders better
    for placeholder in placeholders:
        escaped_placeholder = re.escape('{' + placeholder + '}')
        # Create a named capture group for each placeholder
        # Use a non-greedy match that stops at the next literal character
        regex_pattern = regex_pattern.replace(
            escaped_placeholder, f'(?P<{placeholder}>.*?)'
        )

    # For the last placeholder, we might want greedy matching
    # Adjust the pattern to make the last group greedy if it's at the end
    if regex_pattern.endswith('.*?)'):
        regex_pattern = regex_pattern[:-4] + '.*)'

    # Add anchors to ensure full string match
    regex_pattern = f'^{regex_pattern}$'

    try:
        match = re.match(regex_pattern, string)
        if not match:
            raise absorb.NameParseError(
                f"String '{string}' does not match template '{template}'"
            )

        return match.groupdict()
    except re.error as e:
        raise absorb.NameParseError(f'Invalid regex pattern: {e}')


def convert_raw_parameter_types(
    raw_parameters: dict[str, str],
    parameter_types: dict[str, type | tuple[type, ...]],
) -> dict[str, typing.Any]:
    parameters = {}
    for key, value in raw_parameters.items():
        if key == 'class_name':
            parameter_type: type | tuple[type, ...] = str
        else:
            parameter_type = parameter_types[key]
        if parameter_type == str:  # noqa: E721
            converted: typing.Any = value
        elif parameter_type == int:  # noqa: E721
            converted = int(value)
        elif parameter_type == list[str]:
            converted = value.split(',')
        elif parameter_type == list[int]:
            converted = [int(subvalue) for subvalue in value.split(',')]
        else:
            raise Exception('invalid parameter type: ' + str(parameter_type))
        parameters[key] = converted
    return parameters



from __future__ import annotations

import typing

import absorb


def get_available_range(
    dataset: absorb.TableReference,
) -> absorb.Coverage | None:
    table = absorb.Table.instantiate(dataset)
    return table.get_available_range()


def get_collected_range(
    dataset: absorb.TableReference,
) -> absorb.Coverage | None:
    table = absorb.Table.instantiate(dataset)
    return table.get_collected_range()


def get_collected_tables() -> list[absorb.TableDict]:
    import json
    import os

    datasets_dir = absorb.ops.get_datasets_dir()
    if not os.path.isdir(datasets_dir):
        return []

    tables = []
    for source in os.listdir(datasets_dir):
        for table in os.listdir(absorb.ops.get_source_tables_dir(source)):
            path = absorb.ops.get_table_metadata_path(table, source=source)
            with open(path) as f:
                table_data = json.load(f)
            tables.append(table_data)
    return tables


def get_untracked_collected_tables(
    *, tracked_datasets: list[absorb.TableDict] | None = None
) -> list[absorb.TableDict]:
    import json

    if tracked_datasets is None:
        tracked_datasets = absorb.ops.get_tracked_tables()
    hashed_tracked_datasets = {
        json.dumps(dataset, sort_keys=True) for dataset in tracked_datasets
    }
    return [
        dataset
        for dataset in get_collected_tables()
        if json.dumps(dataset, sort_keys=True) not in hashed_tracked_datasets
    ]



from __future__ import annotations

import typing

import absorb

if typing.TYPE_CHECKING:
    import polars as pl


class Treasuries(absorb.Table):
    source = 'tic'
    description = 'Holdings of US Treasury securities per each country'
    url = 'https://home.treasury.gov/data/treasury-international-capital-tic-system'
    write_range = 'overwrite_all'
    index_type = 'month'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Date,
            'country': pl.String,
            'for_treas_pos': pl.Float64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        return get_post_2019_holdings()

    def get_available_range(self) -> absorb.Coverage:
        import datetime

        holdings = get_post_2019_holdings()
        return (datetime.datetime(2011, 9, 1), holdings['timestamp'].max())


def get_post_2019_holdings() -> pl.DataFrame:
    import io
    import requests
    import polars as pl

    # TODO: get pre-2019 data from https://treasury.gov/resource-center/data-chart-center/tic/Documents/slt3d_globl.csv  # noqa

    url = 'https://ticdata.treasury.gov/resource-center/data-chart-center/tic/Documents/slt_table3.txt'  # noqa
    response = requests.get(url)
    end_index = response.text.index('Notes:	')
    file = io.StringIO(response.text[:end_index])

    schema = {
        'country': pl.String,
        'country_code': pl.String,
        'timestamp': pl.String,
        'for_treas_pos': pl.String,
        'for_treas_net': pl.String,
        'for_lt_treas_pos': pl.String,
        'for_lt_treas_net': pl.String,
        'for_lt_treas_valchg': pl.String,
        'for_st_treas_pos': pl.String,
        'for_st_treas_net': pl.String,
    }

    holders = (
        pl.read_csv(
            file,
            separator='\t',
            schema=schema,
            skip_rows=10,
            truncate_ragged_lines=True,
        )[:-1]
        .with_columns(pl.col('*').replace('n.a.', None))
        .with_columns(
            country_code=pl.col.country_code.cast(int),
            timestamp=pl.col.timestamp.str.to_date('%Y-%m'),
            for_treas_pos=pl.col.for_treas_pos.cast(float),
            for_treas_net=pl.col.for_treas_net.cast(float),
            for_lt_treas_pos=pl.col.for_lt_treas_pos.cast(float),
            for_lt_treas_net=pl.col.for_lt_treas_net.cast(float),
            for_lt_treas_valchg=pl.col.for_lt_treas_valchg.cast(float),
            for_st_treas_pos=pl.col.for_st_treas_pos.cast(float),
            for_st_treas_net=pl.col.for_st_treas_net.cast(float),
        )
        .select(
            'timestamp',
            'country',
            'for_treas_pos',
        )
        .sort('country', 'timestamp')
    )

    return holders



from __future__ import annotations

import typing

import absorb
from . import common


if typing.TYPE_CHECKING:
    import polars as pl


class Stablecoins(absorb.Table):
    source = 'defillama'
    description = 'Total circulating stablecoins in USD'
    url = 'https://defillama.com/'
    write_range = 'overwrite_all'
    index_type = 'day'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {'timestamp': pl.Float64, 'circulating_usd': pl.Float64}

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        return get_historical_total_stablecoins()


class StablecoinsOfChains(absorb.Table):
    source = 'defillama'
    description = 'Circulating stablecoins per chain in USD'
    url = 'https://defillama.com/'
    write_range = 'overwrite_all'
    parameter_types = {'chains': (list, type(None))}
    default_parameters = {'chains': None}
    index_type = 'day'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Float64,
            'chain': pl.String,
            'circulating_usd': pl.Float64,
            'minted_usd': pl.Float64,
            'bridged_usd': pl.Float64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import polars as pl

        chains = self.parameters['chains']
        if chains is None:
            chains = _get_stablecoin_chains()

        dfs = []
        print('collecting', len(chains), 'chains')
        for c, chain in enumerate(chains, start=1):
            print('[' + str(c) + ' / ' + str(len(chains)) + ']', chain)
            try:
                df = get_historical_stablecoins_of_chain(chain)
                dfs.append(df)
            except Exception:
                print('could not collect', chain)
        return pl.concat(dfs)


class StablecoinsOfTokens(absorb.Table):
    source = 'defillama'
    description = 'Circulating stablecoins per token and per chain in USD'
    url = 'https://defillama.com/'
    write_range = 'overwrite_all'
    parameter_types = {'tokens': (list, type(None))}
    default_parameters = {'tokens': None}
    index_type = 'day'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Datetime(time_unit='ms', time_zone=None),
            'token': pl.String,
            'chain': pl.String,
            'circulating': pl.Float64,
            'unreleased': pl.Float64,
            'minted': pl.Float64,
            'bridged_to': pl.Float64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import polars as pl

        tokens = self.parameters['tokens']
        if tokens is None:
            tokens = _get_stablecoin_tokens()

        dfs = []
        print('collecting', len(tokens), 'tokens')
        for t, token in enumerate(tokens, start=1):
            print('[' + str(t) + ' / ' + str(len(tokens)) + ']', token)
            df = get_historical_stablecoins_per_chain_of_token(token)
            dfs.append(df)
        return pl.concat(dfs)


class StablecoinPrices(absorb.Table):
    source = 'defillama'
    description = 'Prices of stablecoins in USD over time'
    url = 'https://defillama.com/'
    write_range = 'overwrite_all'
    index_type = 'day'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Float64,
            'token': pl.String,
            'price': pl.Float64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        return get_historical_stablecoin_prices()


#
# # historical stablecoin getters
#


def _get_stablecoin_chains() -> list[str]:
    data = common._fetch('current_stablecoins')
    return sorted(chain['name'] for chain in data['chains'])


def _get_stablecoin_tokens() -> list[str]:
    data = common._fetch('current_stablecoins')
    return sorted(asset['id'] for asset in data['peggedAssets'])


def get_historical_total_stablecoins() -> pl.DataFrame:
    import polars as pl

    data = common._fetch('historical_total_stablecoins')
    rows = [
        [datum['date'], sum(datum['totalCirculatingUSD'].values())]
        for datum in data
    ]
    schema = {'timestamp': pl.Float64, 'circulating_usd': pl.Float64}
    return pl.DataFrame(rows, schema=schema, orient='row').with_columns(
        (pl.col.timestamp.cast(float) * 1000).cast(pl.Datetime('ms'))
    )


def get_historical_stablecoins_of_chain(chain: str) -> pl.DataFrame:
    import polars as pl

    data = common._fetch('historical_stablecoins_of_chain', {'chain': chain})

    # timestamp, circulating_usd, minted_usd, bridged_usd, chain
    rows = [
        [
            datum['date'],
            chain,
            sum(datum['totalCirculatingUSD'].values()),
            sum(datum.get('totalMintedUSD', {'': 0}).values()),
            sum(datum.get('totalBridgedToUSD', {'': 0}).values()),
        ]
        for datum in data
    ]
    schema = {
        'timestamp': pl.Float64,
        'chain': pl.String,
        'circulating_usd': pl.Float64,
        'minted_usd': pl.Float64,
        'bridged_Usd': pl.Float64,
    }
    return pl.DataFrame(rows, schema=schema, orient='row').with_columns(
        (pl.col.timestamp.cast(float) * 1000).cast(pl.Datetime('ms'))
    )


def get_historical_stablecoins_of_token(token: str) -> pl.DataFrame:
    import polars as pl

    data = common._fetch('historical_stablecoins_of_token', {'token': token})
    rows = [
        [datum['date'], data['symbol'], datum['circulating'][data['pegType']]]
        for datum in data['tokens']
    ]
    schema = {
        'timestamp': pl.Float64,
        'token': pl.String,
        'circulating': pl.Float64,
    }
    return pl.DataFrame(rows, schema=schema, orient='row').with_columns(
        (pl.col.timestamp.cast(float) * 1000).cast(pl.Datetime('ms'))
    )


def get_historical_stablecoins_per_chain_of_token(token: str) -> pl.DataFrame:
    import polars as pl

    data = common._fetch('historical_stablecoins_of_token', {'token': token})
    balances = data['chainBalances']
    peg_type = data['pegType']
    empty = {peg_type: 0}
    rows = [
        [
            datum['date'],
            data['symbol'],
            chain,
            datum.get('circulating', empty).get(peg_type, 0),
            datum.get('unreleased', empty).get(peg_type, 0),
            datum.get('minted', empty).get(peg_type, 0),
            datum.get('bridgedTo', empty).get(peg_type, 0),
        ]
        for chain in balances.keys()
        for datum in balances[chain]['tokens']
    ]
    schema = {
        'timestamp': pl.Float64,
        'token': pl.String,
        'chain': pl.String,
        'circulating': pl.Float64,
        'unreleased': pl.Float64,
        'minted': pl.Float64,
        'bridged_to': pl.Float64,
    }

    return (
        pl.DataFrame(rows, schema=schema, orient='row')
        .with_columns(
            (pl.col.timestamp.cast(float) * 1000).cast(pl.Datetime('ms'))
        )
        .sort('timestamp')
    )


def get_historical_stablecoin_prices() -> pl.DataFrame:
    import polars as pl

    data = common._fetch('historical_stablecoin_prices')
    rows = [
        [datum['date'], str(token), float(price)]
        for datum in data
        for token, price in datum['prices'].items()
    ]
    schema = {'timestamp': pl.Float64, 'token': pl.String, 'price': pl.Float64}
    return (
        pl.DataFrame(rows, schema=schema, orient='row')
        .filter(pl.col.timestamp > 0)
        .with_columns((pl.col.timestamp * 1000).cast(pl.Datetime('ms')))
    )


#
# # current stablecoin getters
#


def get_current_stablecoin_summary(
    *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    import polars as pl

    if data is None:
        data = common._fetch('current_stablecoins')
    rows = []
    for asset in data['peggedAssets']:
        row = dict(asset)
        row['circulating'] = row['circulating'][row['pegType']]
        rows.append(row)
    return pl.DataFrame(rows).drop(
        'circulatingPrevDay',
        'circulatingPrevWeek',
        'circulatingPrevMonth',
        'chainCirculating',
        'chains',
    )


def get_current_stablecoins_per_chain_per_token(
    *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    import polars as pl

    if data is None:
        data = common._fetch('current_stablecoins')
    rows = []
    for asset in data['peggedAssets']:
        for chain in asset['chainCirculating'].keys():
            row = dict(asset)
            row['chain'] = chain
            row['circulating'] = asset['chainCirculating'][chain]['current'][
                row['pegType']
            ]
            rows.append(row)
    return pl.DataFrame(rows).drop(
        'circulatingPrevDay',
        'circulatingPrevWeek',
        'circulatingPrevMonth',
        'chainCirculating',
        'chains',
    )


def get_current_stablecoins_per_chain(
    *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    import polars as pl

    if data is None:
        data = common._fetch('current_stablecoins')
    rows = []
    for chain in data['chains']:
        row = dict(chain)
        rows.append(row)
    return pl.DataFrame(rows).select('name', tvl_usd='tvl').sort('name')


def get_current_stablecoins_per_chain_per_peg_type(
    *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    import polars as pl

    if data is None:
        data = common._fetch('current_stablecoins')
    rows = []
    for chain in data['chains']:
        for peg, amount in chain['totalCirculatingUSD'].items():
            row = dict(chain)
            row['peg_type'] = peg
            row['circulating'] = amount
            rows.append(row)
    return (
        pl.DataFrame(rows)
        .select(
            'name',
            peg_type=pl.col.peg_type.str.strip_prefix('pegged'),
            circulating_native='circulating',
            tvl_usd='tvl',
        )
        .sort('name')
    )



from __future__ import annotations

import typing
import absorb

if typing.TYPE_CHECKING:
    import datetime
    from typing import TypeVar, Protocol

    class SupportsComparison(Protocol):
        def __lt__(self, other: object) -> bool: ...
        def __le__(self, other: object) -> bool: ...
        def __gt__(self, other: object) -> bool: ...
        def __ge__(self, other: object) -> bool: ...
        def __eq__(self, other: object) -> bool: ...

    # _T = TypeVar('_T', int, datetime.datetime, bound=SupportsComparison)
    _T = TypeVar('_T', bound=SupportsComparison)


def index_is_temporal(index_type: absorb.IndexType | None) -> bool:
    return index_type in [
        'hour',
        'day',
        'week',
        'month',
        'quarter',
        'year',
        'timestamp_range',
    ]


def coverage_to_list(
    coverage: absorb.Coverage,
    index_type: absorb.IndexType,
) -> absorb.ChunkList:
    if isinstance(coverage, list):
        if all(isinstance(item, tuple) for item in coverage) and index_type in [
            'hour',
            'day',
            'week',
            'month',
            'quarter',
            'year',
        ]:
            return [
                subitem
                for item in coverage
                for subitem in coverage_to_list(item, index_type)
            ]
        else:
            return coverage
    elif isinstance(coverage, dict):
        if not isinstance(index_type, dict):
            raise Exception()
        if index_type['type'] == 'multi':
            return _multi_coverage_to_list(coverage, index_type)
        else:
            raise NotImplementedError(
                'using number_range or timestamp_range with interval size'
            )
    elif isinstance(coverage, tuple):
        import tooltime

        start, end = coverage
        if index_type in ['hour', 'day', 'week', 'month', 'quarter', 'year']:
            if not isinstance(index_type, str):  # for mypy
                raise Exception()
            return tooltime.get_intervals(
                start,
                end,
                interval=index_type,
                include_end=True,
            )['start'].to_list()
        elif index_type == 'number':
            return list(range(start, end + 1))
        else:
            raise Exception('cannot use this chunk_type as tuple range')
    else:
        raise Exception('invalid coverage format')


def _multi_coverage_to_list(
    coverage: absorb.Coverage,
    index_type: absorb.MultiIndexType,
) -> absorb.ChunkList:
    import itertools

    if not isinstance(coverage, dict):
        raise Exception()
    keys = list(coverage.keys())
    dims = [
        coverage_to_list(
            coverage=coverage[key],
            index_type=index_type['dims'][key],
        )
        for key in keys
    ]
    return [dict(zip(keys, combo)) for combo in itertools.product(*dims)]


def get_range_diff(
    subtract_this: absorb.Coverage,
    from_this: absorb.Coverage,
    index_type: absorb.IndexType,
) -> absorb.Coverage:
    """
    subtraction behaves differently depending on range format
    - mainly, index_type is discrete-closed or continuous-semiopen or other
    - some of these cases will have equivalent outcomes
        - handling them separately keeps maximum clarity + robustness

                                           fs         fe
    original interval                      |----------|
    16 cases of subtraction    1.  |----|
                               2.  |-------|
                               3.  |------------|
                               4.  |------------------|
                               5.  |------------------------|
                               6.          |
                               7.          |------|
                               8.          |----------|
                               9.          |---------------|
                               10.             |
                               11.             |----|
                               12.             |------|
                               13.             |-----------|
                               14.                    |
                               15.                    |-----|
                               16.                        |----|
                                                          ss   se

    if fs == fe
                                            |
                                1.    |--|
                                2.    |-----|
                                3.    |--------|
                                4.          |
                                5.          |--|
                                6.             |--|
    """
    non_range_types = [
        'number_list',
        'timestamp',
        'id',
        'id_range',
        'id_list',
    ]

    if (
        isinstance(subtract_this, (list, dict))
        or isinstance(from_this, (list, dict))
        or index_type in non_range_types
    ):
        if not isinstance(subtract_this, list):
            subtract_this = coverage_to_list(subtract_this, index_type)
        if not isinstance(from_this, list):
            from_this = coverage_to_list(from_this, index_type)
        return [item for item in from_this if item not in subtract_this]

    if not isinstance(subtract_this, tuple) or not isinstance(from_this, tuple):
        raise Exception()
    if index_type in [
        'hour',
        'day',
        'week',
        'month',
        'quarter',
        'year',
    ]:
        import datetime
        import tooltime

        # get discrete chunk
        discrete_step: datetime.timedelta | tooltime.DateDelta
        if index_type == 'hour':
            discrete_step = datetime.timedelta(hours=1)
        elif index_type == 'day':
            discrete_step = datetime.timedelta(days=1)
        elif index_type == 'week':
            discrete_step = datetime.timedelta(days=7)
        elif index_type == 'month':
            discrete_step = tooltime.DateDelta(months=1)
        elif index_type == 'quarter':
            discrete_step = tooltime.DateDelta(quarters=1)
        elif index_type == 'year':
            discrete_step = tooltime.DateDelta(years=1)
        else:
            raise Exception('invalid index_type')

        range_list: absorb.Coverage = _get_discrete_closed_range_diff(
            subtract_this=subtract_this,
            from_this=from_this,
            discrete_step=discrete_step,
        )
    elif index_type == 'timestamp_range':
        range_list = _get_continuous_closed_open_range_diff(
            subtract_this=subtract_this,
            from_this=from_this,
        )
    elif index_type == 'number':
        range_list = _get_discrete_closed_range_diff(
            subtract_this=subtract_this,
            from_this=from_this,
            discrete_step=1,
        )
    elif index_type == 'number_range':
        range_list = _get_discrete_closed_range_diff(
            subtract_this=subtract_this,
            from_this=from_this,
            discrete_step=1,
        )
    elif isinstance(index_type, absorb.CustomIndexType):
        raise NotImplementedError()
    else:
        raise Exception('invalid index_type')

    if len(range_list) == 0:
        return []
    elif len(range_list) == 1:
        return range_list[0]  # type: ignore
    else:
        return range_list
        # return [
        #     item
        #     for range in range_list
        #     for item in coverage_to_list(range, index_type=index_type)
        # ]


def _get_discrete_closed_range_diff(
    subtract_this: tuple[_T, _T],
    from_this: tuple[_T, _T],
    discrete_step: typing.Any,
) -> list[tuple[_T, _T]]:
    s_start, s_end = subtract_this
    f_start, f_end = from_this

    # validity checks
    if s_start > s_end:
        raise Exception('invalid interval, start must be <= end')
    if f_start > f_end:
        raise Exception('invalid interval, start must be <= end')

    # 6 possible cases when f_start == f_end
    if f_start == f_end:
        if s_start < f_start and s_end < f_start:
            # case 1
            return [(f_start, f_end)]
        elif s_start < f_start and s_end == f_start:
            # case 2
            return []
        elif s_start < f_start and s_end > f_start:
            # case 3
            return []
        elif s_start == f_start and s_end == f_start:
            # case 4
            return []
        elif s_start == f_start and s_end > f_start:
            # case 5
            return []
        elif s_start > f_start and s_end > f_start:
            # case 6
            return [(f_start, f_end)]
        else:
            raise Exception()

    # 16 possible cases when f_start < f_end
    if s_start < f_start and s_end < f_start:
        # case 1
        return [(f_start, f_end)]
    elif s_start < f_start and s_end == f_start:
        # case 2
        return [(s_end + discrete_step, f_end)]
    elif s_start < f_start and s_end < f_end:
        # case 3
        return [(s_end + discrete_step, f_end)]
    elif s_start < f_start and s_end == f_end:
        # case 4
        return []
    elif s_start < f_start and s_end > f_end:
        # case 5
        return []
    elif s_start == f_start and s_end == f_start:
        # case 6
        return [(s_end + discrete_step, f_end)]
    elif s_start == f_start and s_end < f_end:
        # case 7
        return [(s_end + discrete_step, f_end)]
    elif s_start == f_start and s_end == f_end:
        # case 8
        return []
    elif s_start == f_start and s_end > f_end:
        # case 9
        return []
    elif s_start < f_end and s_end == s_start:
        # case 10
        return [
            (f_start, s_start - discrete_step),
            (s_end + discrete_step, f_end),
        ]
    elif s_start < f_end and s_end < f_end:
        # case 11
        return [
            (f_start, s_start - discrete_step),
            (s_end + discrete_step, f_end),
        ]
    elif s_start < f_end and s_end == f_end:
        # case 12
        return [(f_start, s_start - discrete_step)]
    elif s_start < f_end and s_end > f_end:
        # case 13
        return [(f_start, s_start - discrete_step)]
    elif s_start == f_end and s_end == f_end:
        # case 14
        return [(f_start, s_start - discrete_step)]
    elif s_start == f_end and s_end > f_end:
        # case 15
        return [(f_start, s_start - discrete_step)]
    elif s_start > f_end and s_end > f_start:
        # case 16
        return [(f_start, f_end)]
    else:
        raise Exception()


def _get_continuous_closed_open_range_diff(
    subtract_this: tuple[_T, _T], from_this: tuple[_T, _T]
) -> list[tuple[_T, _T]]:
    s_start, s_end = subtract_this
    f_start, f_end = from_this

    # validity checks
    if s_start >= s_end:
        raise Exception('invalid interval, start must be < end')
    if f_start >= f_end:
        raise Exception('invalid interval, start must be < end')

    # 16 possible cases
    if s_start < f_start and s_end < f_start:
        # case 1
        return [(f_start, f_end)]
    elif s_start < f_start and s_end == f_start:
        # case 2
        return [(f_start, f_end)]
    elif s_start < f_start and s_end < f_end:
        # case 3
        return [(s_end, f_end)]
    elif s_start < f_start and s_end == f_end:
        # case 4
        return []
    elif s_start < f_start and s_end > f_end:
        # case 5
        return []
    elif s_start == f_start and s_end == f_start:
        # case 6
        raise Exception('s_start should not equal s_end')
    elif s_start == f_start and s_end < f_end:
        # case 7
        return [(s_end, f_end)]
    elif s_start == f_start and s_end == f_end:
        # case 8
        return []
    elif s_start == f_start and s_end > f_end:
        # case 9
        return []
    elif s_start < f_end and s_end == s_start:
        # case 10
        raise Exception('s_start should not equal s_end')
    elif s_start < f_end and s_end < f_end:
        # case 11
        return [(f_start, s_start), (s_end, f_end)]
    elif s_start < f_end and s_end == f_end:
        # case 12
        return [(f_start, s_start)]
    elif s_start < f_end and s_end > f_end:
        # case 13
        return [(f_start, s_start)]
    elif s_start == f_end and s_end == f_end:
        # case 14
        raise Exception('s_start should not equal s_end')
    elif s_start == f_end and s_end > f_end:
        # case 15
        return [(f_start, f_end)]
    elif s_start > f_end and s_end > f_start:
        # case 16
        return [(f_start, f_end)]
    else:
        raise Exception()


def partition_into_chunks(
    coverage: absorb.Coverage, index_type: absorb.IndexType
) -> absorb.ChunkList:
    return coverage_to_list(coverage=coverage, index_type=index_type)
